@@ -1,40 +1,65 @@
+// SPDX-FileCopyrightText: 2000-2002 Michael R. Elkins <me@mutt.org>
+// SPDX-FileCopyrightText: 2002-2022 Oswald Buddenhagen <ossi@users.sf.net>
+// SPDX-License-Identifier: GPL-2.0-or-later WITH LicenseRef-isync-GPL-exception
 /*
  * mbsync - mailbox synchronizer
- * Copyright (C) 2000-2002 Michael R. Elkins <me@mutt.org>
- * Copyright (C) 2002-2006,2011 Oswald Buddenhagen <ossi@users.sf.net>
- *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
- * As a special exception, mbsync may be linked with the OpenSSL library,
- * despite that library's more restrictive license.
  */
 
+#define DEBUG_FLAG DEBUG_MAIN
+
 #include "config.h"
 
 #include "sync.h"
 
-#include <assert.h>
-#include <unistd.h>
-#include <limits.h>
 #include <pwd.h>
-#include <sys/types.h>
 #include <ctype.h>
-#include <string.h>
-#include <stdlib.h>
-#include <stdio.h>
+#include <errno.h>
+#include <sys/stat.h>
 
-static store_conf_t *stores;
+#if defined(WIN32) || defined(_WIN32) || defined(__WIN32__) || defined(__NT__) || defined(__CYGWIN__)
+char FieldDelimiter = ';';
+#else
+char FieldDelimiter = ':';
+#endif
+
+DEF_BIT_FORMATTER_FUNCTION(ops, sync_op_enum)
+
+char *
+expand_strdup( const char *s, const conffile_t *cfile )
+{
+	struct passwd *pw;
+	const char *p, *q;
+	char *r;
+
+	if (*s == '~') {
+		s++;
+		if (!*s) {
+			p = NULL;
+			q = Home;
+		} else if (*s == '/') {
+			p = s;
+			q = Home;
+		} else {
+			if ((p = strchr( s, '/' ))) {
+				r = nfstrndup( s, (size_t)(p - s) );
+				pw = getpwnam( r );
+				free( r );
+			} else {
+				pw = getpwnam( s );
+			}
+			if (!pw)
+				return NULL;
+			q = pw->pw_dir;
+		}
+		nfasprintf( &r, "%s%s", q, p ? p : "" );
+		return r;
+	} else if (*s != '/') {
+		nfasprintf( &r, "%.*s%s", cfile->path_len, cfile->file, s );
+		return r;
+	} else {
+		return nfstrdup( s );
+	}
+}
 
 char *
 get_arg( conffile_t *cfile, int required, int *comment )
@@ -61,14 +86,15 @@ get_arg( conffile_t *cfile, int required, int *comment )
 			if (escaped && c >= 32) {
 				escaped = 0;
 				*t++ = c;
-			} else if (c == '\\')
+			} else if (c == '\\') {
 				escaped = 1;
-			else if (c == '"')
+			} else if (c == '"') {
 				quoted ^= 1;
-			else if (!quoted && isspace( (uchar)c ))
+			} else if (!quoted && isspace( (uchar)c )) {
 				break;
-			else
+			} else {
 				*t++ = c;
+			}
 		}
 		*t = 0;
 		if (escaped) {
@@ -148,6 +174,7 @@ static const struct {
 	const char *name;
 } boxOps[] = {
 	{ OP_EXPUNGE, "Expunge" },
+	{ OP_EXPUNGE_SOLO, "ExpungeSolo" },
 	{ OP_CREATE, "Create" },
 	{ OP_REMOVE, "Remove" },
 };
@@ -160,53 +187,94 @@ getopt_helper( conffile_t *cfile, int *cops, channel_conf_t *conf )
 
 	if (!strcasecmp( "Sync", cfile->cmd )) {
 		arg = cfile->val;
-		do
-			if (!strcasecmp( "Push", arg ))
+		do {
+			if (!strcasecmp( "Push", arg )) {
 				*cops |= XOP_PUSH;
-			else if (!strcasecmp( "Pull", arg ))
+			} else if (!strcasecmp( "Pull", arg )) {
 				*cops |= XOP_PULL;
-			else if (!strcasecmp( "ReNew", arg ))
-				*cops |= OP_RENEW;
-			else if (!strcasecmp( "New", arg ))
+			} else if (!strcasecmp( "Upgrade", arg )) {
+				*cops |= OP_UPGRADE;
+			} else if (!strcasecmp( "ReNew", arg )) {
+				cfile->renew_warn = 1;
+				*cops |= OP_UPGRADE;
+			} else if (!strcasecmp( "New", arg )) {
 				*cops |= OP_NEW;
-			else if (!strcasecmp( "Delete", arg ))
-				*cops |= OP_DELETE;
-			else if (!strcasecmp( "Flags", arg ))
+			} else if (!strcasecmp( "Old", arg )) {
+				*cops |= OP_OLD;
+			} else if (!strcasecmp( "Gone", arg )) {
+				*cops |= OP_GONE;
+			} else if (!strcasecmp( "Delete", arg )) {
+				cfile->delete_warn = 1;
+				*cops |= OP_GONE;
+			} else if (!strcasecmp( "Flags", arg )) {
 				*cops |= OP_FLAGS;
-			else if (!strcasecmp( "PullReNew", arg ))
-				conf->ops[N] |= OP_RENEW;
-			else if (!strcasecmp( "PullNew", arg ))
+			} else if (!strcasecmp( "All", arg ) || !strcasecmp( "Full", arg )) {
+				*cops |= OP_MASK_TYPE;
+			} else if (!strcasecmp( "PullUpgrade", arg )) {
+				conf->ops[N] |= OP_UPGRADE;
+			} else if (!strcasecmp( "PullReNew", arg )) {
+				cfile->renew_warn = 1;
+				conf->ops[N] |= OP_UPGRADE;
+			} else if (!strcasecmp( "PullNew", arg )) {
 				conf->ops[N] |= OP_NEW;
-			else if (!strcasecmp( "PullDelete", arg ))
-				conf->ops[N] |= OP_DELETE;
-			else if (!strcasecmp( "PullFlags", arg ))
+			} else if (!strcasecmp( "PullOld", arg )) {
+				conf->ops[N] |= OP_OLD;
+			} else if (!strcasecmp( "PullGone", arg )) {
+				conf->ops[N] |= OP_GONE;
+			} else if (!strcasecmp( "PullDelete", arg )) {
+				cfile->delete_warn = 1;
+				conf->ops[N] |= OP_GONE;
+			} else if (!strcasecmp( "PullFlags", arg )) {
 				conf->ops[N] |= OP_FLAGS;
-			else if (!strcasecmp( "PushReNew", arg ))
-				conf->ops[F] |= OP_RENEW;
-			else if (!strcasecmp( "PushNew", arg ))
+			} else if (!strcasecmp( "PullFull", arg )) {
+				conf->ops[N] |= OP_MASK_TYPE;
+			} else if (!strcasecmp( "PushUpgrade", arg )) {
+				conf->ops[F] |= OP_UPGRADE;
+			} else if (!strcasecmp( "PushReNew", arg )) {
+				cfile->renew_warn = 1;
+				conf->ops[F] |= OP_UPGRADE;
+			} else if (!strcasecmp( "PushNew", arg )) {
 				conf->ops[F] |= OP_NEW;
-			else if (!strcasecmp( "PushDelete", arg ))
-				conf->ops[F] |= OP_DELETE;
-			else if (!strcasecmp( "PushFlags", arg ))
+			} else if (!strcasecmp( "PushOld", arg )) {
+				conf->ops[F] |= OP_OLD;
+			} else if (!strcasecmp( "PushGone", arg )) {
+				conf->ops[F] |= OP_GONE;
+			} else if (!strcasecmp( "PushDelete", arg )) {
+				cfile->delete_warn = 1;
+				conf->ops[F] |= OP_GONE;
+			} else if (!strcasecmp( "PushFlags", arg )) {
 				conf->ops[F] |= OP_FLAGS;
-			else if (!strcasecmp( "All", arg ) || !strcasecmp( "Full", arg ))
-				*cops |= XOP_PULL|XOP_PUSH;
-			else if (strcasecmp( "None", arg ) && strcasecmp( "Noop", arg )) {
+			} else if (!strcasecmp( "PushFull", arg )) {
+				conf->ops[F] |= OP_MASK_TYPE;
+			} else if (!strcasecmp( "None", arg ) || !strcasecmp( "Noop", arg )) {
+				conf->ops[F] |= XOP_TYPE_NOOP;
+			} else {
 				error( "%s:%d: invalid Sync arg '%s'\n",
 				       cfile->file, cfile->line, arg );
 				cfile->err = 1;
 			}
-		while ((arg = get_arg( cfile, ARG_OPTIONAL, NULL )));
+		} while ((arg = get_arg( cfile, ARG_OPTIONAL, NULL )));
 		conf->ops[F] |= XOP_HAVE_TYPE;
-	} else if (!strcasecmp( "SyncState", cfile->cmd ))
-		conf->sync_state = expand_strdup( cfile->val );
-	else if (!strcasecmp( "CopyArrivalDate", cfile->cmd ))
+	} else if (!strcasecmp( "SyncState", cfile->cmd )) {
+		conf->sync_state = !strcmp( cfile->val, "*" ) ? "*" : expand_strdup( cfile->val, cfile );
+	} else if (!strcasecmp( "CopyArrivalDate", cfile->cmd )) {
 		conf->use_internal_date = parse_bool( cfile );
-	else if (!strcasecmp( "MaxMessages", cfile->cmd ))
+	} else if (!strcasecmp( "MaxMessages", cfile->cmd )) {
 		conf->max_messages = parse_int( cfile );
-	else if (!strcasecmp( "ExpireUnread", cfile->cmd ))
+	} else if (!strcasecmp( "ExpireSide", cfile->cmd )) {
+		arg = cfile->val;
+		if (!strcasecmp( "Far", arg )) {
+			conf->expire_side = F;
+		} else if (!strcasecmp( "Near", arg )) {
+			conf->expire_side = N;
+		} else {
+			error( "%s:%d: invalid ExpireSide argument '%s'\n",
+			       cfile->file, cfile->line, arg );
+			cfile->err = 1;
+		}
+	} else if (!strcasecmp( "ExpireUnread", cfile->cmd )) {
 		conf->expire_unread = parse_bool( cfile );
-	else {
+	} else {
 		for (i = 0; i < as(boxOps); i++) {
 			if (!strcasecmp( boxOps[i].name, cfile->cmd )) {
 				int op = boxOps[i].op;
@@ -224,7 +292,9 @@ getopt_helper( conffile_t *cfile, int *cops, channel_conf_t *conf )
 					} else if (!strcasecmp( "Slave", arg )) {  // Pre-1.4 legacy
 						conf->ops[N] |= op;
 						cfile->ms_warn = 1;
-					} else if (strcasecmp( "None", arg )) {
+					} else if (!strcasecmp( "None", arg )) {
+						conf->ops[F] |= op * (XOP_EXPUNGE_NOOP / OP_EXPUNGE);
+					} else {
 						error( "%s:%d: invalid %s arg '%s'\n",
 						       cfile->file, cfile->line, boxOps[i].name, arg );
 						cfile->err = 1;
@@ -264,38 +334,76 @@ getcline( conffile_t *cfile )
 	return 0;
 }
 
-/* XXX - this does not detect None conflicts ... */
+static const char *
+channel_str( const char *chan_name )
+{
+	if (!chan_name)
+		return "on the command line";
+
+	if (!*chan_name)
+		return "in global config section";
+
+	static char buf[100];
+	nfsnprintf( buf, sizeof(buf), "in Channel '%s'", chan_name );
+	return buf;
+}
+
 int
-merge_ops( int cops, int ops[] )
+merge_ops( int cops, int ops[], const char *chan_name )
 {
 	int aops, op;
 	uint i;
 
+	if (!cops && !ops[F] && !ops[N])  // Only to denoise the debug output
+		return 0;
+	debug( "merge ops (%s):\n  common: %s\n  far: %s\n  near: %s\n",
+	       channel_str( chan_name ), fmt_ops( cops ).str, fmt_ops( ops[F] ).str, fmt_ops( ops[N] ).str );
 	aops = ops[F] | ops[N];
 	if (ops[F] & XOP_HAVE_TYPE) {
-		if (aops & OP_MASK_TYPE) {
-			if (aops & cops & OP_MASK_TYPE) {
+		if (aops & OP_MASK_TYPE) {  // PullNew, etc.
+			if (ops[F] & XOP_TYPE_NOOP) {
 			  cfl:
-				error( "Conflicting Sync args specified.\n" );
+				error( "Conflicting Sync options specified %s.\n", channel_str( chan_name ) );
 				return 1;
 			}
-			ops[F] |= cops & OP_MASK_TYPE;
-			ops[N] |= cops & OP_MASK_TYPE;
+			if (aops & cops & OP_MASK_TYPE) {  // Overlapping New, etc.
+			  ovl:
+				error( "Redundant Sync options specified %s.\n", channel_str( chan_name ) );
+				return 1;
+			}
+			// Mix in non-overlapping Push/Pull or New, etc.
 			if (cops & XOP_PULL) {
-				if (ops[N] & OP_MASK_TYPE)
-					goto cfl;
-				ops[N] |= OP_MASK_TYPE;
+				if (cops & (XOP_PUSH | OP_MASK_TYPE)) {
+					// Mixing instant effect flags with row/column flags would be confusing,
+					// so instead everything is instant effect. This implies that mixing
+					// direction with type would cause overlaps, so PullNew Push Gone, etc.
+					// is invalid.
+					// Pull Push covers everything, so makes no sense to combine.
+				  ivl:
+					error( "Invalid combination of simple and compound Sync options %s.\n",
+					       channel_str( chan_name ) );
+					return 1;
+				}
+				if (ops[N] & OP_DFLT_TYPE)
+					goto ovl;
+				ops[N] |= OP_DFLT_TYPE;
+			} else if (cops & XOP_PUSH) {
+				if (cops & OP_MASK_TYPE)
+					goto ivl;
+				if (ops[F] & OP_DFLT_TYPE)
+					goto ovl;
+				ops[F] |= OP_DFLT_TYPE;
+			} else {
+				ops[F] |= cops & OP_MASK_TYPE;
+				ops[N] |= cops & OP_MASK_TYPE;
 			}
-			if (cops & XOP_PUSH) {
-				if (ops[F] & OP_MASK_TYPE)
-					goto cfl;
-				ops[F] |= OP_MASK_TYPE;
-			}
-		} else if (cops & (OP_MASK_TYPE|XOP_MASK_DIR)) {
+		} else if (cops & (OP_MASK_TYPE | XOP_MASK_DIR)) {  // Pull New, etc.
+			if (ops[F] & XOP_TYPE_NOOP)
+				goto cfl;
 			if (!(cops & OP_MASK_TYPE))
-				cops |= OP_MASK_TYPE;
+				cops |= OP_DFLT_TYPE;
 			else if (!(cops & XOP_MASK_DIR))
-				cops |= XOP_PULL|XOP_PUSH;
+				cops |= XOP_PULL | XOP_PUSH;
 			if (cops & XOP_PULL)
 				ops[N] |= cops & OP_MASK_TYPE;
 			if (cops & XOP_PUSH)
@@ -305,14 +413,19 @@ merge_ops( int cops, int ops[] )
 	for (i = 0; i < as(boxOps); i++) {
 		op = boxOps[i].op;
 		if (ops[F] & (op * (XOP_HAVE_EXPUNGE / OP_EXPUNGE))) {
+			if (((aops | cops) & op) && (ops[F] & (op * (XOP_EXPUNGE_NOOP / OP_EXPUNGE)))) {
+				error( "Conflicting %s options specified %s.\n", boxOps[i].name, channel_str( chan_name ) );
+				return 1;
+			}
 			if (aops & cops & op) {
-				error( "Conflicting %s args specified.\n", boxOps[i].name );
+				error( "Redundant %s options specified %s.\n", boxOps[i].name, channel_str( chan_name ) );
 				return 1;
 			}
 			ops[F] |= cops & op;
 			ops[N] |= cops & op;
 		}
 	}
+	debug( "  => far: %s\n  => near: %s\n", fmt_ops( ops[F] ).str, fmt_ops( ops[N] ).str );
 	return 0;
 }
 
@@ -327,14 +440,39 @@ load_config( const char *where )
 	char *arg, *p;
 	uint len, max_size;
 	int cops, gcops, glob_ok, fn, i;
-	char path[_POSIX_PATH_MAX];
+	char path[_POSIX_PATH_MAX], path2[_POSIX_PATH_MAX];
 	char buf[1024];
 
 	if (!where) {
-		nfsnprintf( path, sizeof(path), "%s/." EXE "rc", Home );
-		cfile.file = path;
-	} else
-		cfile.file = where;
+		int path_len, path_len2;
+		const char *config_home = getenv( "XDG_CONFIG_HOME" );
+		if (config_home)
+			nfsnprintf( path, sizeof(path), "%s/%nisyncrc", config_home, &path_len );
+		else
+			nfsnprintf( path, sizeof(path), "%s/.config/%nisyncrc", Home, &path_len );
+		nfsnprintf( path2, sizeof(path2), "%s/%n.mbsyncrc", Home, &path_len2 );
+		struct stat st;
+		int ex = !lstat( path, &st );
+		int ex2 = !lstat( path2, &st );
+		if (ex2 && !ex) {
+			cfile.file = path2;
+			cfile.path_len = path_len2;
+		} else {
+			if (ex && ex2)
+				warn( "Both %s and %s exist; using the former.\n", path, path2 );
+			cfile.file = path;
+			cfile.path_len = path_len;
+		}
+	} else {
+		const char *sl = strrchr( where, '/' );
+		if (!sl) {
+			nfsnprintf( path, sizeof(path), "./%n%s", &cfile.path_len, where );
+			cfile.file = path;
+		} else {
+			cfile.path_len = sl - where + 1;
+			cfile.file = where;
+		}
+	}
 
 	info( "Reading configuration file %s\n", cfile.file );
 
@@ -348,16 +486,19 @@ load_config( const char *where )
 	cfile.line = 0;
 	cfile.err = 0;
 	cfile.ms_warn = 0;
+	cfile.renew_warn = 0;
+	cfile.delete_warn = 0;
 	cfile.rest = NULL;
 
 	gcops = 0;
 	glob_ok = 1;
+	global_conf.expire_side = N;
 	global_conf.expire_unread = -1;
   reloop:
 	while (getcline( &cfile )) {
 		if (!cfile.cmd)
 			continue;
-		for (i = 0; i < N_DRIVERS; i++)
+		for (i = 0; i < N_DRIVERS; i++) {
 			if (drivers[i]->parse_store( &cfile, &store )) {
 				if (store) {
 					if (!store->max_size)
@@ -371,27 +512,26 @@ load_config( const char *where )
 				glob_ok = 0;
 				goto reloop;
 			}
-		if (!strcasecmp( "Channel", cfile.cmd ))
-		{
-			channel = nfcalloc( sizeof(*channel) );
+		}
+		if (!strcasecmp( "Channel", cfile.cmd )) {
+			channel = nfzalloc( sizeof(*channel) );
 			channel->name = nfstrdup( cfile.val );
 			channel->max_messages = global_conf.max_messages;
+			channel->expire_side = global_conf.expire_side;
 			channel->expire_unread = global_conf.expire_unread;
 			channel->use_internal_date = global_conf.use_internal_date;
 			cops = 0;
 			max_size = UINT_MAX;
 			while (getcline( &cfile ) && cfile.cmd) {
-				if (!strcasecmp( "MaxSize", cfile.cmd ))
+				if (!strcasecmp( "MaxSize", cfile.cmd )) {
 					max_size = parse_size( &cfile );
-				else if (!strcasecmp( "Pattern", cfile.cmd ) ||
-				         !strcasecmp( "Patterns", cfile.cmd ))
-				{
+				} else if (!strcasecmp( "Pattern", cfile.cmd ) ||
+				           !strcasecmp( "Patterns", cfile.cmd )) {
 					arg = cfile.val;
-					do
+					do {
 						add_string_list( &channel->patterns, arg );
-					while ((arg = get_arg( &cfile, ARG_OPTIONAL, NULL )));
-				}
-				else if (!strcasecmp( "Far", cfile.cmd )) {
+					} while ((arg = get_arg( &cfile, ARG_OPTIONAL, NULL )));
+				} else if (!strcasecmp( "Far", cfile.cmd )) {
 					fn = F;
 					goto linkst;
 				} else if (!strcasecmp( "Master", cfile.cmd )) {  // Pre-1.4 legacy
@@ -412,11 +552,13 @@ load_config( const char *where )
 						continue;
 					}
 					*p = 0;
-					for (store = stores; store; store = store->next)
+					for (store = stores; store; store = store->next) {
 						if (!strcmp( store->name, cfile.val + 1 )) {
 							channel->stores[fn] = store;
 							goto stpcom;
 						}
+					}
+					channel->stores[fn] = (void *)~0;
 					error( "%s:%d: unknown store '%s'\n",
 					       cfile.file, cfile.line, cfile.val + 1 );
 					cfile.err = 1;
@@ -427,31 +569,30 @@ load_config( const char *where )
 				} else if (!getopt_helper( &cfile, &cops, channel )) {
 					error( "%s:%d: keyword '%s' is not recognized in Channel sections\n",
 					       cfile.file, cfile.line, cfile.cmd );
+					cfile.rest = NULL;
 					cfile.err = 1;
 				}
 			}
 			if (!channel->stores[F]) {
 				error( "channel '%s' refers to no far side store\n", channel->name );
 				cfile.err = 1;
-			} else if (!channel->stores[N]) {
+			}
+			if (!channel->stores[N]) {
 				error( "channel '%s' refers to no near side store\n", channel->name );
 				cfile.err = 1;
-			} else if (merge_ops( cops, channel->ops ))
-				cfile.err = 1;
-			else {
-				if (max_size != UINT_MAX) {
-					if (!max_size)
-						max_size = UINT_MAX;
-					channel->stores[F]->max_size = channel->stores[N]->max_size = max_size;
-				}
-				*channelapp = channel;
-				channelapp = &channel->next;
 			}
+			if (merge_ops( cops, channel->ops, channel->name ))
+				cfile.err = 1;
+			if (max_size != UINT_MAX && !cfile.err) {
+				if (!max_size)
+					max_size = UINT_MAX;
+				channel->stores[F]->max_size = channel->stores[N]->max_size = max_size;
+			}
+			*channelapp = channel;
+			channelapp = &channel->next;
 			glob_ok = 0;
 			goto reloop;
-		}
-		else if (!strcasecmp( "Group", cfile.cmd ))
-		{
+		} else if (!strcasecmp( "Group", cfile.cmd )) {
 			group = nfmalloc( sizeof(*group) );
 			group->name = nfstrdup( cfile.val );
 			*groupapp = group;
@@ -470,27 +611,21 @@ load_config( const char *where )
 			}
 			while (getcline( &cfile ) && cfile.cmd) {
 				if (!strcasecmp( "Channel", cfile.cmd ) ||
-				    !strcasecmp( "Channels", cfile.cmd ))
-				{
+				    !strcasecmp( "Channels", cfile.cmd )) {
 					arg = cfile.val;
 					goto addone;
-				}
-				else
-				{
+				} else {
 					error( "%s:%d: keyword '%s' is not recognized in Group sections\n",
 					       cfile.file, cfile.line, cfile.cmd );
+					cfile.rest = NULL;
 					cfile.err = 1;
 				}
 			}
 			glob_ok = 0;
 			goto reloop;
-		}
-		else if (!strcasecmp( "FSync", cfile.cmd ))
-		{
+		} else if (!strcasecmp( "FSync", cfile.cmd )) {
 			UseFSync = parse_bool( &cfile );
-		}
-		else if (!strcasecmp( "FieldDelimiter", cfile.cmd ))
-		{
+		} else if (!strcasecmp( "FieldDelimiter", cfile.cmd )) {
 			if (strlen( cfile.val ) != 1) {
 				error( "%s:%d: Field delimiter must be exactly one character long\n", cfile.file, cfile.line );
 				cfile.err = 1;
@@ -501,20 +636,17 @@ load_config( const char *where )
 					cfile.err = 1;
 				}
 			}
-		}
-		else if (!strcasecmp( "BufferLimit", cfile.cmd ))
-		{
+		} else if (!strcasecmp( "BufferLimit", cfile.cmd )) {
 			BufferLimit = parse_size( &cfile );
 			if (!BufferLimit) {
 				error( "%s:%d: BufferLimit cannot be zero\n", cfile.file, cfile.line );
 				cfile.err = 1;
 			}
-		}
-		else if (!getopt_helper( &cfile, &gcops, &global_conf ))
-		{
+		} else if (!getopt_helper( &cfile, &gcops, &global_conf )) {
 			error( "%s:%d: '%s' is not a recognized section-starting or global keyword\n",
 			       cfile.file, cfile.line, cfile.cmd );
 			cfile.err = 1;
+			cfile.rest = NULL;
 			while (getcline( &cfile ))
 				if (!cfile.cmd)
 					goto reloop;
@@ -529,8 +661,30 @@ load_config( const char *where )
 	fclose (cfile.fp);
 	if (cfile.ms_warn)
 		warn( "Notice: Master/Slave are deprecated; use Far/Near instead.\n" );
-	cfile.err |= merge_ops( gcops, global_conf.ops );
-	if (!global_conf.sync_state)
-		global_conf.sync_state = expand_strdup( "~/." EXE "/" );
+	if (cfile.renew_warn)
+		warn( "Notice: ReNew is deprecated; use Upgrade instead.\n" );
+	if (cfile.delete_warn)
+		warn( "Notice: Delete is deprecated; use Gone instead.\n" );
+	cfile.err |= merge_ops( gcops, global_conf.ops, "" );
+	if (!global_conf.sync_state) {
+		const char *state_home = getenv( "XDG_STATE_HOME" );
+		if (state_home)
+			nfsnprintf( path, sizeof(path), "%s/isync/", state_home );
+		else
+			nfsnprintf( path, sizeof(path), "%s/.local/state/isync/", Home );
+		nfsnprintf( path2, sizeof(path2), "%s/.mbsync/", Home );
+		struct stat st;
+		int ex = !lstat( path, &st );
+		int ex2 = !lstat( path2, &st );
+		if (ex2 && !ex) {
+			global_conf.sync_state = nfstrdup( path2 );
+		} else {
+			if (ex && ex2) {
+				error( "Error: both %s and %s exist; delete one or set SyncState globally.\n", path, path2 );
+				cfile.err = 1;
+			}
+			global_conf.sync_state = nfstrdup( path );
+		}
+	}
 	return cfile.err;
 }
