@@ -1,50 +1,68 @@
+// SPDX-FileCopyrightText: 2000-2002 Michael R. Elkins <me@mutt.org>
+// SPDX-FileCopyrightText: 2002-2022 Oswald Buddenhagen <ossi@users.sf.net>
+// SPDX-License-Identifier: GPL-2.0-or-later WITH LicenseRef-isync-GPL-exception
 /*
  * mbsync - mailbox synchronizer
- * Copyright (C) 2000-2002 Michael R. Elkins <me@mutt.org>
- * Copyright (C) 2002-2006,2010-2012 Oswald Buddenhagen <ossi@users.sf.net>
- *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
- * As a special exception, mbsync may be linked with the OpenSSL library,
- * despite that library's more restrictive license.
  */
 
 #include "driver.h"
 
-#include <stdlib.h>
-#include <string.h>
+store_conf_t *stores;
 
 driver_t *drivers[N_DRIVERS] = { &maildir_driver, &imap_driver };
 
-uint
-count_generic_messages( message_t *msgs )
+void
+cleanup_drivers( void )
 {
-	uint count = 0;
-	for (; msgs; msgs = msgs->next)
-		count++;
-	return count;
+	for (int t = 0; t < N_DRIVERS; t++)
+		drivers[t]->cleanup();
+}
+
+// Keep the MESSAGE_FLAGS in sync (grep that)!
+const char MsgFlags[] = { 'D', 'F', 'P', 'R', 'S', 'T' };
+
+static void
+make_flags( uchar flags, char *buf )
+{
+	uint i, d;
+
+	for (i = d = 0; i < as(MsgFlags); i++)
+		if (flags & (1 << i))
+			buf[d++] = MsgFlags[i];
+	buf[d] = 0;
+}
+
+flag_str_t
+fmt_flags( uchar flags )
+{
+	flag_str_t buf;
+
+	make_flags( flags, buf.str );
+	return buf;
+}
+
+flag_str_t
+fmt_lone_flags( uchar flags )
+{
+	flag_str_t buf;
+
+	if (!flags) {
+		buf.str[0] = '-';
+		buf.str[1] = 0;
+	} else {
+		make_flags( flags, buf.str );
+	}
+	return buf;
 }
 
 void
 free_generic_messages( message_t *msgs )
 {
-	message_t *tmsg;
-
-	for (; msgs; msgs = tmsg) {
-		tmsg = msgs->next;
+	while (msgs) {
+		message_t *tmsg = msgs->next;
 		free( msgs->msgid );
 		free( msgs );
+		msgs = tmsg;
 	}
 }
 
@@ -73,6 +91,7 @@ parse_generic_store( store_conf_t *store, conffile_t *cfg, const char *type )
 		store->flat_delim = nfstrdup( cfg->val );
 	} else {
 		error( "%s:%d: keyword '%s' is not recognized in %s sections\n", cfg->file, cfg->line, cfg->cmd, type );
+		cfg->rest = NULL;
 		cfg->err = 1;
 	}
 }
