@@ -1,40 +1,20 @@
+// SPDX-FileCopyrightText: 2000-2002 Michael R. Elkins <me@mutt.org>
+// SPDX-FileCopyrightText: 2002-2022 Oswald Buddenhagen <ossi@users.sf.net>
+// SPDX-FileCopyrightText: 2004 Theodore Y. Ts'o <tytso@mit.edu>
+// SPDX-License-Identifier: GPL-2.0-or-later WITH LicenseRef-isync-GPL-exception
 /*
  * mbsync - mailbox synchronizer
- * Copyright (C) 2000-2002 Michael R. Elkins <me@mutt.org>
- * Copyright (C) 2002-2006,2010-2013 Oswald Buddenhagen <ossi@users.sf.net>
- * Copyright (C) 2004 Theodore Y. Ts'o <tytso@mit.edu>
- *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
- * As a special exception, mbsync may be linked with the OpenSSL library,
- * despite that library's more restrictive license.
  */
 
+#define DEBUG_FLAG DEBUG_MAILDIR
+
 #include "driver.h"
 
-#include <assert.h>
-#include <limits.h>
-#include <stdlib.h>
-#include <string.h>
 #include <ctype.h>
 #include <dirent.h>
 #include <fcntl.h>
-#include <stdio.h>
-#include <unistd.h>
 #include <sys/stat.h>
 #include <errno.h>
-#include <time.h>
 #include <utime.h>
 
 #if !defined(_POSIX_SYNCHRONIZED_IO) || _POSIX_SYNCHRONIZED_IO <= 0
@@ -55,6 +35,7 @@ typedef union maildir_store_conf {
 	struct {
 		STORE_CONF
 		char *path;
+		char *path_sfx;
 		char *inbox;
 #ifdef USE_DB
 		int alt_map;
@@ -93,10 +74,12 @@ typedef union maildir_store {
 		// but mailbox totals. also, don't trust them beyond the initial load.
 		int total_msgs, recent_msgs;
 		maildir_message_t *msgs;
+		maildir_message_t **app_msgs;  // Only for testing find_new()
 		wakeup_t lcktmr;
 
+		void (*expunge_callback)( message_t *msg, void *aux );
 		void (*bad_callback)( void *aux );
-		void *bad_callback_aux;
+		void *callback_aux;
 	};
 } maildir_store_t;
 
@@ -107,21 +90,6 @@ static struct flock lck;
 
 static int MaildirCount;
 
-static void ATTR_PRINTFLIKE(1, 2)
-debug( const char *msg, ... )
-{
-	va_list va;
-
-	va_start( va, msg );
-	vdebug( DEBUG_SYNC, msg, va );
-	va_end( va );
-}
-
-/* Keep the mailbox driver flag definitions in sync: */
-/* grep for MAILBOX_DRIVER_FLAG */
-/* The order is according to alphabetical maildir flag sort */
-static const char Flags[] = { 'D', 'F', 'P', 'R', 'S', 'T' };
-
 static uchar
 maildir_parse_flags( const char *info_prefix, const char *base )
 {
@@ -131,8 +99,8 @@ maildir_parse_flags( const char *info_prefix, const char *base )
 
 	flags = 0;
 	if ((s = strstr( base, info_prefix )))
-		for (s += 3, i = 0; i < as(Flags); i++)
-			if (strchr( s, Flags[i] ))
+		for (s += 3, i = 0; i < as(MsgFlags); i++)
+			if (strchr( s, MsgFlags[i] ))
 				flags |= (1 << i);
 	return flags;
 }
@@ -154,19 +122,23 @@ static char *
 maildir_join_path( maildir_store_conf_t *conf, int in_inbox, const char *box )
 {
 	char *out, *p;
-	const char *prefix;
-	uint pl, bl, n;
+	const char *prefix, *infix;
+	uint pl, il, bl, n;
 	char c;
 
 	if (in_inbox || conf->sub_style == SUB_MAILDIRPP) {
 		prefix = conf->inbox;
+		infix = NULL;
+		il = 0;
 	} else {
 		if (maildir_ensure_path( conf ) < 0)
 			return NULL;
 		prefix = conf->path;
+		infix = conf->path_sfx;
+		il = strlen( infix ) + 1;
 	}
 	pl = strlen( prefix );
-	for (bl = 0, n = 0; (c = box[bl]); bl++)
+	for (bl = 0, n = 0; (c = box[bl]); bl++) {
 		if (c == '/') {
 			if (conf->sub_style == SUB_UNSET) {
 				error( "Maildir error: accessing subfolder '%s', but store '%s' does not specify SubFolders style\n",
@@ -179,6 +151,7 @@ maildir_join_path( maildir_store_conf_t *conf, int in_inbox, const char *box )
 			       conf->name, box );
 			return NULL;
 		}
+	}
 	switch (conf->sub_style) {
 	case SUB_VERBATIM:
 		n = 0;
@@ -189,12 +162,16 @@ maildir_join_path( maildir_store_conf_t *conf, int in_inbox, const char *box )
 	default: /* SUB_LEGACY and SUB_UNSET */
 		break;
 	}
-	out = nfmalloc( pl + bl + n + 1 );
+	out = nfmalloc( pl + il + bl + n + 1 );
 	memcpy( out, prefix, pl );
 	p = out + pl;
 	if (conf->sub_style == SUB_MAILDIRPP) {
 		*p++ = '/';
 		*p++ = '.';
+	} else if (il) {
+		*p++ = '/';
+		memcpy( p, infix, il - 1 );
+		p += il - 1;
 	}
 	while ((c = *box++)) {
 		if (c == '/') {
@@ -237,7 +214,7 @@ maildir_alloc_store( store_conf_t *gconf, const char *label ATTR_UNUSED )
 {
 	maildir_store_t *ctx;
 
-	ctx = nfcalloc( sizeof(*ctx) );
+	ctx = nfzalloc( sizeof(*ctx) );
 	ctx->driver = &maildir_driver;
 	ctx->gen.conf = gconf;
 	ctx->uvfd = -1;
@@ -266,10 +243,12 @@ maildir_connect_store( store_t *gctx,
 static void
 free_maildir_messages( maildir_message_t *msg )
 {
-	for (maildir_message_t *tmsg; (tmsg = msg); msg = tmsg) {
-		tmsg = msg->next;
+	while (msg) {
+		maildir_message_t *tmsg = msg->next;
 		free( msg->base );
+		free( msg->msgid );
 		free( msg );
+		msg = tmsg;
 	}
 }
 
@@ -309,18 +288,20 @@ maildir_cleanup_drv( void )
 }
 
 static void
-maildir_set_bad_callback( store_t *gctx, void (*cb)( void *aux ), void *aux )
+maildir_set_callbacks( store_t *gctx, void (*exp_cb)( message_t *, void * ),
+                       void (*bad_cb)( void * ), void *aux )
 {
 	maildir_store_t *ctx = (maildir_store_t *)gctx;
 
-	ctx->bad_callback = cb;
-	ctx->bad_callback_aux = aux;
+	ctx->expunge_callback = exp_cb;
+	ctx->bad_callback = bad_cb;
+	ctx->callback_aux = aux;
 }
 
 static void
 maildir_invoke_bad_callback( maildir_store_t *ctx )
 {
-	ctx->bad_callback( ctx->bad_callback_aux );
+	ctx->bad_callback( ctx->callback_aux );
 }
 
 static int
@@ -391,12 +372,10 @@ maildir_list_maildirpp( maildir_store_t *ctx, int flags, const char *inbox )
 	return 0;
 }
 
-static int maildir_list_inbox( maildir_store_t *ctx, int flags, const char *basePath );
-static int maildir_list_path( maildir_store_t *ctx, int flags, const char *inbox );
-
 static int
-maildir_list_recurse( maildir_store_t *ctx, int isBox, int flags, int depth,
-                      const char *inbox, uint inboxLen, const char *basePath, uint basePathLen,
+maildir_list_recurse( maildir_store_t *ctx, int isBox,
+                      const char *inbox, uint inboxLen,
+                      char *suffix, int suffixLen,
                       char *path, int pathLen, char *name, int nameLen )
 {
 	DIR *dir;
@@ -417,7 +396,7 @@ maildir_list_recurse( maildir_store_t *ctx, int isBox, int flags, int depth,
 		closedir( dir );
 		return -1;
 	}
-	if (++depth > 10) {
+	if (isBox > 10) {
 		// We do the other checks first to avoid confusing error messages for files.
 		error( "Maildir error: path %s is too deeply nested. Symlink loop?\n", path );
 		closedir( dir );
@@ -430,19 +409,19 @@ maildir_list_recurse( maildir_store_t *ctx, int isBox, int flags, int depth,
 		pl = nfsnprintf( path + pathLen, _POSIX_PATH_MAX - pathLen, "%s", ent );
 		if (pl == 3 && (!memcmp( ent, "cur", 3 ) || !memcmp( ent, "new", 3 ) || !memcmp( ent, "tmp", 3 )))
 			continue;
+		if (suffixLen) {
+			if (!starts_with( ent, pl, suffix, suffixLen ))
+				continue;
+			if (pl == suffixLen) {
+				error( "Maildir error: empty mailbox name under %s - did you forget the trailing slash?\n", path );
+				closedir( dir );
+				return -1;
+			}
+			ent += suffixLen;
+		}
 		pl += pathLen;
 		if (inbox && equals( path, pl, inbox, inboxLen )) {
 			// Inbox nested into Path.
-			if (maildir_list_inbox( ctx, flags, NULL ) < 0) {
-				closedir( dir );
-				return -1;
-			}
-		} else if (basePath && equals( path, pl, basePath, basePathLen )) {
-			// Path nested into Inbox.
-			if (maildir_list_path( ctx, flags, NULL ) < 0) {
-				closedir( dir );
-				return -1;
-			}
 		} else {
 			if (style == SUB_LEGACY) {
 				if (*ent == '.') {
@@ -470,7 +449,7 @@ maildir_list_recurse( maildir_store_t *ctx, int isBox, int flags, int depth,
 				add_string_list( &ctx->boxes, name );
 			path[pl] = 0;
 			name[nl++] = '/';
-			if (maildir_list_recurse( ctx, isBox + 1, flags, depth, inbox, inboxLen, basePath, basePathLen, path, pl, name, nl ) < 0) {
+			if (maildir_list_recurse( ctx, isBox + 1, inbox, inboxLen, NULL, 0, path, pl, name, nl ) < 0) {
 				closedir( dir );
 				return -1;
 			}
@@ -481,7 +460,7 @@ maildir_list_recurse( maildir_store_t *ctx, int isBox, int flags, int depth,
 }
 
 static int
-maildir_list_inbox( maildir_store_t *ctx, int flags, const char *basePath )
+maildir_list_inbox( maildir_store_t *ctx )
 {
 	char path[_POSIX_PATH_MAX], name[_POSIX_PATH_MAX];
 
@@ -491,13 +470,13 @@ maildir_list_inbox( maildir_store_t *ctx, int flags, const char *basePath )
 
 	add_string_list( &ctx->boxes, "INBOX" );
 	return maildir_list_recurse(
-	        ctx, 1, flags, 0, NULL, 0, basePath, basePath ? strlen( basePath ) - 1 : 0,
+	        ctx, 1, NULL, 0, NULL, 0,
 	        path, nfsnprintf( path, _POSIX_PATH_MAX, "%s/", ctx->conf->inbox ),
 	        name, nfsnprintf( name, _POSIX_PATH_MAX, "INBOX/" ) );
 }
 
 static int
-maildir_list_path( maildir_store_t *ctx, int flags, const char *inbox )
+maildir_list_path( maildir_store_t *ctx )
 {
 	char path[_POSIX_PATH_MAX], name[_POSIX_PATH_MAX];
 
@@ -507,9 +486,11 @@ maildir_list_path( maildir_store_t *ctx, int flags, const char *inbox )
 
 	if (maildir_ensure_path( ctx->conf ) < 0)
 		return -1;
+	const char *inbox = ctx->conf->inbox;
 	return maildir_list_recurse(
-	        ctx, 0, flags, 0, inbox, inbox ? strlen( inbox ) : 0, NULL, 0,
-	        path, nfsnprintf( path, _POSIX_PATH_MAX, "%s", ctx->conf->path ),
+	        ctx, 0, inbox, strlen( inbox ),
+	        ctx->conf->path_sfx, strlen( ctx->conf->path_sfx ),
+	        path, nfsnprintf( path, _POSIX_PATH_MAX, "%s/", ctx->conf->path ),
 	        name, 0 );
 }
 
@@ -523,9 +504,9 @@ maildir_list_store( store_t *gctx, int flags,
 	if (conf->sub_style == SUB_MAILDIRPP
 	        ? maildir_list_maildirpp( ctx, flags, conf->inbox ) < 0
 	        : ((((flags & LIST_PATH) || ((flags & LIST_PATH_MAYBE) && conf->path))
-	            && maildir_list_path( ctx, flags, conf->inbox ) < 0) ||
+	            && maildir_list_path( ctx ) < 0) ||
 	           ((flags & LIST_INBOX)
-	            && maildir_list_inbox( ctx, flags, conf->path ) < 0))) {
+	            && maildir_list_inbox( ctx ) < 0))) {
 		maildir_invoke_bad_callback( ctx );
 		cb( DRV_CANCELED, NULL, aux );
 	} else {
@@ -552,8 +533,10 @@ maildir_free_scan( msg_t_array_alloc_t *msglist )
 	uint i;
 
 	if (msglist->array.data) {
-		for (i = 0; i < msglist->array.size; i++)
+		for (i = 0; i < msglist->array.size; i++) {
 			free( msglist->array.data[i].base );
+			free( msglist->array.data[i].msgid );
+		}
 		free( msglist->array.data );
 	}
 }
@@ -593,21 +576,6 @@ maildir_clear_tmp( char *buf, int bufsz, int bl )
 	return DRV_OK;
 }
 
-static int
-make_box_dir( char *buf, int bl )
-{
-	char *p;
-
-	if (!mkdir( buf, 0700 ) || errno == EEXIST)
-		return 0;
-	p = memrchr( buf, '/', (size_t)bl - 1 );
-	*p = 0;
-	if (make_box_dir( buf, (int)(p - buf) ))
-		return -1;
-	*p = '/';
-	return mkdir( buf, 0700 );
-}
-
 static int
 maildir_validate( const char *box, int create, maildir_store_t *ctx )
 {
@@ -618,20 +586,20 @@ maildir_validate( const char *box, int create, maildir_store_t *ctx )
 	bl = nfsnprintf( buf, sizeof(buf) - 4, "%s/", box );
 	if (stat( buf, &st )) {
 		if (errno != ENOENT) {
-			sys_error( "Maildir error: cannot access mailbox '%s'", box );
+			sys_error( "Maildir error: cannot access mailbox '%s'", buf );
 			return DRV_BOX_BAD;
 		}
 		if (!create)
 			return DRV_BOX_BAD;
-		if (make_box_dir( buf, bl )) {
-			sys_error( "Maildir error: cannot create mailbox '%s'", box );
+		if (mkdir_p( buf, bl - 1 )) {
+			sys_error( "Maildir error: cannot create mailbox '%s'", buf );
 			ctx->conf->failed = FAIL_FINAL;
 			maildir_invoke_bad_callback( ctx );
 			return DRV_CANCELED;
 		}
 	} else if (!S_ISDIR(st.st_mode)) {
 	  notdir:
-		error( "Maildir error: '%s' is no valid mailbox\n", box );
+		error( "Maildir error: '%s' is no valid mailbox\n", buf );
 		return DRV_BOX_BAD;
 	}
 	for (i = 0; i < 3; i++) {
@@ -702,11 +670,11 @@ maildir_store_uidval( maildir_store_t *ctx )
 		n = sprintf( buf, "%u\n%u\n", ctx->uidvalidity, ctx->nuid );
 		lseek( ctx->uvfd, 0, SEEK_SET );
 		if (write( ctx->uvfd, buf, (uint)n ) != n || ftruncate( ctx->uvfd, n ) || (UseFSync && fdatasync( ctx->uvfd ))) {
-			error( "Maildir error: cannot write UIDVALIDITY.\n" );
+			error( "Maildir error: cannot write UIDVALIDITY in %s\n", ctx->path );
 			return DRV_BOX_BAD;
 		}
 	}
-	conf_wakeup( &ctx->lcktmr, 2 );
+	conf_wakeup( &ctx->lcktmr, 2000 );
 	return DRV_OK;
 }
 
@@ -728,7 +696,7 @@ maildir_init_uidval( maildir_store_t *ctx )
 static int
 maildir_init_uidval_new( maildir_store_t *ctx )
 {
-	notice( "Maildir notice: no UIDVALIDITY, creating new.\n" );
+	notice( "Maildir notice: no UIDVALIDITY in %s, creating new.\n", ctx->path );
 	return maildir_init_uidval( ctx );
 }
 
@@ -753,14 +721,14 @@ maildir_uidval_lock( maildir_store_t *ctx )
 #endif
 	lck.l_type = F_WRLCK;
 	if (fcntl( ctx->uvfd, F_SETLKW, &lck )) {
-		error( "Maildir error: cannot fcntl lock UIDVALIDITY.\n" );
+		error( "Maildir error: cannot fcntl lock UIDVALIDITY in %s.\n", ctx->path );
 		return DRV_BOX_BAD;
 	}
 
 #ifdef USE_DB
 	if (ctx->usedb) {
 		if (fstat( ctx->uvfd, &st )) {
-			sys_error( "Maildir error: cannot fstat UID database" );
+			sys_error( "Maildir error: cannot fstat UID database in %s", ctx->path );
 			return DRV_BOX_BAD;
 		}
 		if (db_create( &ctx->db, NULL, 0 )) {
@@ -794,7 +762,7 @@ maildir_uidval_lock( maildir_store_t *ctx )
 			 * But this would mess up the sync state completely. So better bail out and
 			 * give the user a chance to fix the mailbox. */
 			if (n) {
-				error( "Maildir error: cannot read UIDVALIDITY.\n" );
+				error( "Maildir error: cannot read UIDVALIDITY in %s.\n", ctx->path );
 				return DRV_BOX_BAD;
 			}
 #endif
@@ -802,7 +770,7 @@ maildir_uidval_lock( maildir_store_t *ctx )
 		}
 	}
 	ctx->uvok = 1;
-	conf_wakeup( &ctx->lcktmr, 2 );
+	conf_wakeup( &ctx->lcktmr, 2000 );
 	return DRV_OK;
 }
 
@@ -1057,11 +1025,10 @@ maildir_scan( maildir_store_t *ctx, msg_t_array_alloc_t *msglist )
 		}
 #ifdef USE_DB
 		if (ctx->usedb) {
-			if (maildir_uidval_lock( ctx ) != DRV_OK)
-				;
-			else if ((ret = ctx->db->cursor( ctx->db, NULL, &dbc, 0 )))
+			if (maildir_uidval_lock( ctx ) != DRV_OK) {
+			} else if ((ret = ctx->db->cursor( ctx->db, NULL, &dbc, 0 ))) {
 				ctx->db->err( ctx->db, ret, "Maildir error: db->cursor()" );
-			else {
+			} else {
 				for (;;) {
 					if ((ret = dbc->c_get( dbc, &key, &value, DB_NEXT ))) {
 						if (ret != DB_NOTFOUND)
@@ -1092,11 +1059,11 @@ maildir_scan( maildir_store_t *ctx, msg_t_array_alloc_t *msglist )
 				if (uid == entry->uid) {
 #if 1
 					/* See comment in maildir_uidval_lock() why this is fatal. */
-					error( "Maildir error: duplicate UID %u.\n", uid );
+					error( "Maildir error: duplicate UID %u in %s.\n", uid, ctx->path );
 					maildir_free_scan( msglist );
 					return DRV_BOX_BAD;
 #else
-					notice( "Maildir notice: duplicate UID; changing UIDVALIDITY.\n");
+					notice( "Maildir notice: duplicate UID in %s; changing UIDVALIDITY.\n", ctx->path );
 					if ((ret = maildir_init_uid( ctx )) != DRV_OK) {
 						maildir_free_scan( msglist );
 						return ret;
@@ -1109,7 +1076,8 @@ maildir_scan( maildir_store_t *ctx, msg_t_array_alloc_t *msglist )
 				if (uid > ctx->nuid) {
 					/* In principle, we could just warn and top up nuid. However, getting into this
 					 * situation might indicate some serious trouble, so let's not make it worse. */
-					error( "Maildir error: UID %u is beyond highest assigned UID %u.\n", uid, ctx->nuid );
+					error( "Maildir error: UID %u is beyond highest assigned UID %u in %s.\n",
+					       uid, ctx->nuid, ctx->path );
 					maildir_free_scan( msglist );
 					return DRV_BOX_BAD;
 				}
@@ -1153,9 +1121,10 @@ maildir_scan( maildir_store_t *ctx, msg_t_array_alloc_t *msglist )
 				free( entry->base );
 				entry->base = nfstrndup( buf + bl + 4, (size_t)fnl );
 			}
-			int want_size = ((ctx->opts & OPEN_NEW_SIZE) && uid > ctx->newuid);
+			int want_size = ((ctx->opts & OPEN_OLD_SIZE) && uid <= ctx->newuid) ||
+			                ((ctx->opts & OPEN_NEW_SIZE) && uid > ctx->newuid);
 			int want_tuid = ((ctx->opts & OPEN_FIND) && uid >= ctx->finduid);
-			int want_msgid = ((ctx->opts & OPEN_OLD_IDS) && uid <= ctx->pairuid);
+			int want_msgid = ((ctx->opts & OPEN_PAIRED_IDS) && uid <= ctx->pairuid);
 			if (!want_size && !want_tuid && !want_msgid)
 				continue;
 			if (!fnl)
@@ -1191,7 +1160,7 @@ maildir_scan( maildir_store_t *ctx, msg_t_array_alloc_t *msglist )
 						break;
 					if (want_tuid && starts_with( lnbuf, bufl, "X-TUID: ", 8 )) {
 						if (bufl < 8 + TUIDL) {
-							error( "Maildir error: malformed X-TUID header (UID %u)\n", uid );
+							error( "Maildir error: malformed X-TUID header in %s\n", buf );
 							continue;
 						}
 						memcpy( entry->tuid, lnbuf + 8, TUIDL );
@@ -1236,15 +1205,15 @@ maildir_init_msg( maildir_store_t *ctx, maildir_message_t *msg, msg_t *entry )
 	msg->msgid = entry->msgid;
 	entry->msgid = NULL; /* prevent deletion */
 	msg->size = entry->size;
-	msg->srec = NULL;
 	memcpy( msg->tuid, entry->tuid, TUIDL );
 	if (entry->recent)
 		msg->status |= M_RECENT;
 	if (ctx->opts & OPEN_FLAGS) {
 		msg->status |= M_FLAGS;
 		msg->flags = maildir_parse_flags( ctx->conf->info_prefix, msg->base );
-	} else
+	} else {
 		msg->flags = 0;
+	}
 }
 
 static void
@@ -1256,6 +1225,7 @@ maildir_app_msg( maildir_store_t *ctx, maildir_message_t ***msgapp, msg_t *entry
 	*msgapp = &msg->next;
 	msg->uid = entry->uid;
 	msg->status = 0;
+	msg->srec = NULL;
 	maildir_init_msg( ctx, msg, entry );
 }
 
@@ -1267,6 +1237,7 @@ maildir_select_box( store_t *gctx, const char *name )
 
 	maildir_cleanup( gctx );
 	ctx->msgs = NULL;
+	ctx->app_msgs = &ctx->msgs;
 	ctx->excs.data = NULL;
 	ctx->uvfd = -1;
 #ifdef USE_DB
@@ -1309,7 +1280,7 @@ maildir_open_box( store_t *gctx,
 
 	nfsnprintf( uvpath, sizeof(uvpath), "%s/.uidvalidity", ctx->path );
 #ifndef USE_DB
-	if ((ctx->uvfd = open( uvpath, O_RDWR|O_CREAT, 0600 )) < 0) {
+	if ((ctx->uvfd = open( uvpath, O_RDWR | O_CREAT, 0600 )) < 0) {
 		sys_error( "Maildir error: cannot write %s", uvpath );
 		cb( DRV_BOX_BAD, UIDVAL_BAD, aux );
 		return;
@@ -1320,11 +1291,11 @@ maildir_open_box( store_t *gctx,
 		nfsnprintf( uvpath, sizeof(uvpath), "%s/.isyncuidmap.db", ctx->path );
 		if ((ctx->uvfd = open( uvpath, O_RDWR, 0600 )) < 0) {
 			if (ctx->conf->alt_map) {
-				if ((ctx->uvfd = open( uvpath, O_RDWR|O_CREAT, 0600 )) >= 0)
+				if ((ctx->uvfd = open( uvpath, O_RDWR | O_CREAT, 0600 )) >= 0)
 					goto dbok;
 			} else {
 				nfsnprintf( uvpath, sizeof(uvpath), "%s/.uidvalidity", ctx->path );
-				if ((ctx->uvfd = open( uvpath, O_RDWR|O_CREAT, 0600 )) >= 0)
+				if ((ctx->uvfd = open( uvpath, O_RDWR | O_CREAT, 0600 )) >= 0)
 					goto fnok;
 			}
 			sys_error( "Maildir error: cannot write %s", uvpath );
@@ -1442,9 +1413,9 @@ maildir_prepare_load_box( store_t *gctx, uint opts )
 	maildir_store_t *ctx = (maildir_store_t *)gctx;
 
 	if (opts & OPEN_SETFLAGS)
-		opts |= OPEN_OLD;
+		opts |= OPEN_PAIRED;
 	if (opts & OPEN_EXPUNGE)
-		opts |= OPEN_OLD|OPEN_NEW|OPEN_FLAGS;
+		opts |= OPEN_PAIRED | OPEN_OLD | OPEN_NEW | OPEN_FLAGS | OPEN_UID_EXPUNGE;
 	ctx->opts = opts;
 	return opts;
 }
@@ -1466,6 +1437,7 @@ maildir_load_box( store_t *gctx, uint minuid, uint maxuid, uint finduid, uint pa
 	ARRAY_SQUEEZE( &excs );
 	ctx->excs = excs;
 
+	assert( !ctx->msgs );
 	if (maildir_scan( ctx, &msglist ) != DRV_OK) {
 		cb( DRV_BOX_BAD, NULL, 0, 0, aux );
 		return;
@@ -1473,6 +1445,7 @@ maildir_load_box( store_t *gctx, uint minuid, uint maxuid, uint finduid, uint pa
 	msgapp = &ctx->msgs;
 	for (i = 0; i < msglist.array.size; i++)
 		maildir_app_msg( ctx, &msgapp, msglist.array.data + i );
+	ctx->app_msgs = msgapp;
 	maildir_free_scan( &msglist );
 
 	cb( DRV_OK, &ctx->msgs->gen, ctx->total_msgs, ctx->recent_msgs, aux );
@@ -1488,35 +1461,30 @@ maildir_rescan( maildir_store_t *ctx )
 	ctx->fresh[0] = ctx->fresh[1] = 0;
 	if (maildir_scan( ctx, &msglist ) != DRV_OK)
 		return DRV_BOX_BAD;
+	debug( "Maildir processing rescan of %s:\n", ctx->path );
 	for (msgapp = &ctx->msgs, i = 0; (msg = *msgapp) || i < msglist.array.size; ) {
 		if (!msg) {
-#if 0
-			debug( "adding new message %u\n", msglist.array.data[i].uid );
+			debug( "  adding new message %u\n", msglist.array.data[i].uid );
 			maildir_app_msg( ctx, &msgapp, msglist.array.data + i );
-#else
-			debug( "ignoring new message %u\n", msglist.array.data[i].uid );
-#endif
 			i++;
 		} else if (i >= msglist.array.size) {
-			debug( "purging deleted message %u\n", msg->uid );
+			debug( "  purging deleted message %u\n", msg->uid );
 			msg->status = M_DEAD;
+			ctx->expunge_callback( &msg->gen, ctx->callback_aux );
 			msgapp = &msg->next;
 		} else if (msglist.array.data[i].uid < msg->uid) {
 			/* this should not happen, actually */
-#if 0
-			debug( "adding new message %u\n", msglist.array.data[i].uid );
+			debug( "  adding new message %u\n", msglist.array.data[i].uid );
 			maildir_app_msg( ctx, &msgapp, msglist.array.data + i );
-#else
-			debug( "ignoring new message %u\n", msglist.array.data[i].uid );
-#endif
 			i++;
 		} else if (msglist.array.data[i].uid > msg->uid) {
-			debug( "purging deleted message %u\n", msg->uid );
+			debug( "  purging deleted message %u\n", msg->uid );
 			msg->status = M_DEAD;
+			ctx->expunge_callback( &msg->gen, ctx->callback_aux );
 			msgapp = &msg->next;
 		} else {
-			debug( "updating message %u\n", msg->uid );
-			msg->status &= ~(M_FLAGS|M_RECENT);
+			debug( "  updating message %u\n", msg->uid );
+			msg->status &= ~(M_FLAGS | M_RECENT);
 			free( msg->base );
 			free( msg->msgid );
 			maildir_init_msg( ctx, msg, msglist.array.data + i );
@@ -1593,9 +1561,9 @@ maildir_make_flags( char info_delimiter, uchar flags, char *buf )
 	buf[0] = info_delimiter;
 	buf[1] = '2';
 	buf[2] = ',';
-	for (d = 3, i = 0; i < (int)as(Flags); i++)
+	for (d = 3, i = 0; i < (int)as(MsgFlags); i++)
 		if (flags & (1 << i))
-			buf[d++] = Flags[i];
+			buf[d++] = MsgFlags[i];
 	buf[d] = 0;
 	return d;
 }
@@ -1608,7 +1576,7 @@ maildir_store_msg( store_t *gctx, msg_data_t *data, int to_trash,
 	const char *box;
 	int ret, fd, bl;
 	uint uid;
-	char buf[_POSIX_PATH_MAX], nbuf[_POSIX_PATH_MAX], fbuf[NUM_FLAGS + 3], base[128];
+	char buf[_POSIX_PATH_MAX], nbuf[_POSIX_PATH_MAX], fbuf[as(MsgFlags) + 3], base[128];
 
 	bl = nfsnprintf( base, sizeof(base), "%lld.%d_%d.%s", (long long)time( NULL ), Pid, ++MaildirCount, Hostname );
 	if (!to_trash) {
@@ -1637,7 +1605,7 @@ maildir_store_msg( store_t *gctx, msg_data_t *data, int to_trash,
 
 	maildir_make_flags( ctx->conf->info_delimiter, data->flags, fbuf );
 	nfsnprintf( buf, sizeof(buf), "%s/tmp/%s%s", box, base, fbuf );
-	if ((fd = open( buf, O_WRONLY|O_CREAT|O_EXCL, 0600 )) < 0) {
+	if ((fd = open( buf, O_WRONLY | O_CREAT | O_EXCL, 0600 )) < 0) {
 		if (errno != ENOENT || !to_trash) {
 			sys_error( "Maildir error: cannot create %s", buf );
 			free( data->data );
@@ -1649,7 +1617,7 @@ maildir_store_msg( store_t *gctx, msg_data_t *data, int to_trash,
 			cb( ret, 0, aux );
 			return;
 		}
-		if ((fd = open( buf, O_WRONLY|O_CREAT|O_EXCL, 0600 )) < 0) {
+		if ((fd = open( buf, O_WRONLY | O_CREAT | O_EXCL, 0600 )) < 0) {
 			sys_error( "Maildir error: cannot create %s", buf );
 			free( data->data );
 			cb( DRV_BOX_BAD, 0, aux );
@@ -1692,9 +1660,24 @@ maildir_store_msg( store_t *gctx, msg_data_t *data, int to_trash,
 		cb( DRV_BOX_BAD, 0, aux );
 		return;
 	}
+	if (DFlags & FAKEDUMBSTORE)
+		uid = 0;
 	cb( DRV_OK, uid, aux );
 }
 
+static void
+maildir_find_new_msgs( store_t *gctx, uint newuid,
+                       void (*cb)( int sts, message_t *msgs, void *aux ), void *aux )
+{
+	maildir_store_t *ctx = (maildir_store_t *)gctx;
+
+	assert( DFlags & FAKEDUMBSTORE );
+	ctx->opts |= OPEN_FIND;
+	ctx->finduid = newuid;
+	int ret = maildir_rescan( ctx );
+	cb( ret, &(*ctx->app_msgs)->gen, aux );
+}
+
 static void
 maildir_set_msg_flags( store_t *gctx, message_t *gmsg, uint uid ATTR_UNUSED, int add, int del,
                        void (*cb)( int sts, void *aux ), void *aux )
@@ -1713,24 +1696,24 @@ maildir_set_msg_flags( store_t *gctx, message_t *gmsg, uint uid ATTR_UNUSED, int
 	for (;;) {
 		bl = bbl + nfsnprintf( buf + bbl, _POSIX_PATH_MAX - bbl, "%s/", subdirs[gmsg->status & M_RECENT] );
 		ol = strlen( msg->base );
-		if (_POSIX_PATH_MAX - bl < ol + 3 + NUM_FLAGS)
+		if (_POSIX_PATH_MAX - bl < ol + 3 + (int)as(MsgFlags))
 			oob();
 		memcpy( buf + bl, msg->base, (size_t)ol + 1 );
 		memcpy( nbuf + bl, msg->base, (size_t)ol + 1 );
 		if ((s = strstr( nbuf + bl, conf->info_prefix ))) {
 			s += 3;
 			fl = ol - (s - (nbuf + bl));
-			for (i = 0; i < as(Flags); i++) {
-				if ((p = strchr( s, Flags[i] ))) {
+			for (i = 0; i < as(MsgFlags); i++) {
+				if ((p = strchr( s, MsgFlags[i] ))) {
 					if (del & (1 << i)) {
 						memmove( p, p + 1, (size_t)fl - (size_t)(p - s) );
 						fl--;
 					}
 				} else if (add & (1 << i)) {
-					for (j = 0; j < fl && Flags[i] > s[j]; j++);
+					for (j = 0; j < fl && MsgFlags[i] > s[j]; j++);
 					fl++;
 					memmove( s + j + 1, s + j, (size_t)(fl - j) );
-					s[j] = Flags[i];
+					s[j] = MsgFlags[i];
 				}
 			}
 			tl = ol + 3 + fl;
@@ -1808,6 +1791,7 @@ maildir_trash_msg( store_t *gctx, message_t *gmsg,
 	}
 	gmsg->status |= M_DEAD;
 	ctx->total_msgs--;
+	ctx->expunge_callback( gmsg, ctx->callback_aux );
 
 #ifdef USE_DB
 	if (ctx->usedb) {
@@ -1820,7 +1804,7 @@ maildir_trash_msg( store_t *gctx, message_t *gmsg,
 
 static void
 maildir_close_box( store_t *gctx,
-                   void (*cb)( int sts, void *aux ), void *aux )
+                   void (*cb)( int sts, int reported, void *aux ), void *aux )
 {
 	maildir_store_t *ctx = (maildir_store_t *)gctx;
 	maildir_message_t *msg;
@@ -1830,8 +1814,8 @@ maildir_close_box( store_t *gctx,
 	for (;;) {
 		retry = 0;
 		basel = nfsnprintf( buf, sizeof(buf), "%s/", ctx->path );
-		for (msg = ctx->msgs; msg; msg = msg->next)
-			if (!(msg->status & M_DEAD) && (msg->flags & F_DELETED)) {
+		for (msg = ctx->msgs; msg; msg = msg->next) {
+			if (!(msg->status & M_DEAD) && (msg->status & M_EXPUNGE)) {
 				nfsnprintf( buf + basel, _POSIX_PATH_MAX - basel, "%s/%s", subdirs[msg->status & M_RECENT], msg->base );
 				if (unlink( buf )) {
 					if (errno == ENOENT)
@@ -1841,20 +1825,22 @@ maildir_close_box( store_t *gctx,
 				} else {
 					msg->status |= M_DEAD;
 					ctx->total_msgs--;
+					ctx->expunge_callback( &msg->gen, ctx->callback_aux );
 #ifdef USE_DB
 					if (ctx->db && (ret = maildir_purge_msg( ctx, msg->base )) != DRV_OK) {
-						cb( ret, aux );
+						cb( ret, 1, aux );
 						return;
 					}
 #endif /* USE_DB */
 				}
 			}
+		}
 		if (!retry) {
-			cb( DRV_OK, aux );
+			cb( DRV_OK, 1, aux );
 			return;
 		}
-		if ((ret = maildir_rescan( (maildir_store_t *)gctx )) != DRV_OK) {
-			cb( ret, aux );
+		if ((ret = maildir_rescan( ctx )) != DRV_OK) {
+			cb( ret, 1, aux );
 			return;
 		}
 	}
@@ -1867,12 +1853,6 @@ maildir_cancel_cmds( store_t *gctx ATTR_UNUSED,
 	cb( aux );
 }
 
-static void
-maildir_commit_cmds( store_t *gctx )
-{
-	(void) gctx;
-}
-
 static uint
 maildir_get_memory_usage( store_t *gctx ATTR_UNUSED )
 {
@@ -1892,21 +1872,26 @@ maildir_parse_store( conffile_t *cfg, store_conf_t **storep )
 
 	if (strcasecmp( "MaildirStore", cfg->cmd ))
 		return 0;
-	store = nfcalloc( sizeof(*store) );
+	store = nfzalloc( sizeof(*store) );
 	store->info_delimiter = FieldDelimiter;
 	store->driver = &maildir_driver;
 	store->name = nfstrdup( cfg->val );
 
-	while (getcline( cfg ) && cfg->cmd)
-		if (!strcasecmp( "Inbox", cfg->cmd ))
-			store->inbox = expand_strdup( cfg->val );
-		else if (!strcasecmp( "Path", cfg->cmd ))
-			store->path = expand_strdup( cfg->val );
+	while (getcline( cfg ) && cfg->cmd) {
+		if (!strcasecmp( "Inbox", cfg->cmd )) {
+			store->inbox = expand_strdup( cfg->val, cfg );
+		} else if (!strcasecmp( "Path", cfg->cmd )) {
+			store->path = expand_strdup( cfg->val, cfg );
+		} else if (!strcasecmp( "AltMap", cfg->cmd )) {
 #ifdef USE_DB
-		else if (!strcasecmp( "AltMap", cfg->cmd ))
 			store->alt_map = parse_bool( cfg );
+#else
+			if (parse_bool( cfg )) {
+				error( "Error: AltMap=true is not supported by this build.\n" );
+				cfg->err = 1;
+			}
 #endif /* USE_DB */
-		else if (!strcasecmp( "InfoDelimiter", cfg->cmd )) {
+		} else if (!strcasecmp( "InfoDelimiter", cfg->cmd )) {
 			if (strlen( cfg->val ) != 1) {
 				error( "%s:%d: Info delimiter must be exactly one character long\n", cfg->file, cfg->line );
 				cfg->err = 1;
@@ -1929,13 +1914,28 @@ maildir_parse_store( conffile_t *cfg, store_conf_t **storep )
 				error( "%s:%d: Unrecognized SubFolders style\n", cfg->file, cfg->line );
 				cfg->err = 1;
 			}
-		} else
+		} else {
 			parse_generic_store( &store->gen, cfg, "MaildirStore" );
+		}
+	}
 	if (!store->inbox)
-		store->inbox = expand_strdup( "~/Maildir" );
-	if (store->sub_style == SUB_MAILDIRPP && store->path) {
-		error( "Maildir store '%s': Setting Path is incompatible with 'SubFolders Maildir++'\n", store->name );
-		cfg->err = 1;
+		store->inbox = expand_strdup( "~/Maildir", NULL );
+	if (store->path) {
+		if (store->sub_style == SUB_MAILDIRPP) {
+			error( "Maildir store '%s': Setting Path is incompatible with 'SubFolders Maildir++'\n", store->name );
+			cfg->err = 1;
+		} else {
+			uint inboxLen = strlen( store->inbox );
+			if (starts_with( store->path, -1, store->inbox, inboxLen ) && store->path[inboxLen] == '/') {
+				error( "Maildir store '%s': Path cannot be nested under Inbox\n", store->name );
+				cfg->err = 1;
+			} else {
+				char *s = strrchr( store->path, '/' );
+				assert( s );  // due to expand_strdup()
+				store->path_sfx = s + 1;
+				*s = 0;
+			}
+		}
 	}
 	nfasprintf( &store->info_prefix, "%c2,", store->info_delimiter );
 	nfasprintf( &store->info_stop, "%c,", store->info_delimiter );
@@ -1954,7 +1954,7 @@ struct driver maildir_driver = {
 	maildir_parse_store,
 	maildir_cleanup_drv,
 	maildir_alloc_store,
-	maildir_set_bad_callback,
+	maildir_set_callbacks,
 	maildir_connect_store,
 	maildir_free_store,
 	maildir_free_store, /* _cancel_, but it's the same */
@@ -1972,12 +1972,11 @@ struct driver maildir_driver = {
 	maildir_load_box,
 	maildir_fetch_msg,
 	maildir_store_msg,
-	NULL,  // find_new_msgs
+	maildir_find_new_msgs,
 	maildir_set_msg_flags,
 	maildir_trash_msg,
 	maildir_close_box,
 	maildir_cancel_cmds,
-	maildir_commit_cmds,
 	maildir_get_memory_usage,
 	maildir_get_fail_state,
 };
