@@ -1,34 +1,14 @@
+// SPDX-FileCopyrightText: 2000-2002 Michael R. Elkins <me@mutt.org>
+// SPDX-FileCopyrightText: 2002-2022 Oswald Buddenhagen <ossi@users.sf.net>
+// SPDX-FileCopyrightText: 2004 Theodore Y. Ts'o <tytso@mit.edu>
+// SPDX-License-Identifier: GPL-2.0-or-later WITH LicenseRef-isync-GPL-exception
 /*
  * mbsync - mailbox synchronizer
- * Copyright (C) 2000-2002 Michael R. Elkins <me@mutt.org>
- * Copyright (C) 2002-2006,2008,2010,2011, 2013 Oswald Buddenhagen <ossi@users.sf.net>
- * Copyright (C) 2004 Theodore Y. Ts'o <tytso@mit.edu>
- *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
- * As a special exception, mbsync may be linked with the OpenSSL library,
- * despite that library's more restrictive license.
  */
 
 #include "socket.h"
 
-#include <assert.h>
-#include <unistd.h>
-#include <stdlib.h>
-#include <stddef.h>
 #include <errno.h>
-#include <string.h>
 #include <fcntl.h>
 #include <sys/socket.h>
 #include <sys/ioctl.h>
@@ -48,6 +28,7 @@
 #endif
 
 enum {
+	SCK_RESOLVING,
 	SCK_CONNECTING,
 #ifdef HAVE_LIBSSL
 	SCK_STARTTLS,
@@ -402,7 +383,7 @@ socket_start_deflate( conn_t *conn )
 {
 	int result;
 
-	conn->in_z = nfcalloc( sizeof(*conn->in_z) );
+	conn->in_z = nfzalloc( sizeof(*conn->in_z) );
 	result = inflateInit2(
 			conn->in_z,
 			-15 /* Use raw deflate */
@@ -412,7 +393,7 @@ socket_start_deflate( conn_t *conn )
 		abort();
 	}
 
-	conn->out_z = nfcalloc( sizeof(*conn->out_z) );
+	conn->out_z = nfzalloc( sizeof(*conn->out_z) );
 	result = deflateInit2(
 			conn->out_z,
 			Z_DEFAULT_COMPRESSION, /* Compression level */
@@ -427,6 +408,7 @@ socket_start_deflate( conn_t *conn )
 	}
 
 	init_wakeup( &conn->z_fake, z_fake_cb, conn );
+	conn->readsz = 0;  // This optimization makes no sense past this point
 }
 #endif /* HAVE_LIBZ */
 
@@ -434,6 +416,7 @@ static void socket_fd_cb( int, void * );
 static void socket_fake_cb( void * );
 static void socket_timeout_cb( void * );
 
+static void socket_resolve( conn_t * );
 static void socket_connect_one( conn_t * );
 static void socket_connect_next( conn_t * );
 static void socket_connect_failed( conn_t * );
@@ -441,15 +424,21 @@ static void socket_connected( conn_t * );
 static void socket_connect_bail( conn_t * );
 
 static void
-socket_open_internal( conn_t *sock, int fd )
+socket_register_internal( conn_t *sock, int fd )
 {
 	sock->fd = fd;
-	fcntl( fd, F_SETFL, O_NONBLOCK );
 	init_notifier( &sock->notify, fd, socket_fd_cb, sock );
 	init_wakeup( &sock->fd_fake, socket_fake_cb, sock );
 	init_wakeup( &sock->fd_timeout, socket_timeout_cb, sock );
 }
 
+static void
+socket_open_internal( conn_t *sock, int fd )
+{
+	fcntl( fd, F_SETFL, O_NONBLOCK );
+	socket_register_internal( sock, fd );
+}
+
 static void
 socket_close_internal( conn_t *sock )
 {
@@ -460,32 +449,6 @@ socket_close_internal( conn_t *sock )
 	sock->fd = -1;
 }
 
-#ifndef HAVE_IPV6
-struct addr_info {
-	struct addr_info *ai_next;
-	struct sockaddr_in ai_addr[1];
-};
-
-#define freeaddrinfo(ai) free( ai )
-
-static struct addr_info *
-init_addrinfo( struct hostent *he )
-{
-	uint naddr = 0;
-	for (char **addr = he->h_addr_list; *addr; addr++)
-		naddr++;
-	struct addr_info *caddr = nfcalloc( naddr * sizeof(struct addrinfo) );
-	struct addr_info *ret, **caddrp = &ret;
-	for (char **addr = he->h_addr_list; *addr; addr++, caddr++) {
-		caddr->ai_addr->sin_family = AF_INET;
-		memcpy( &caddr->ai_addr->sin_addr.s_addr, *addr, sizeof(struct in_addr) );
-		*caddrp = caddr;
-		caddrp = &caddr->ai_next;
-	}
-	return ret;
-}
-#endif
-
 void
 socket_connect( conn_t *sock, void (*cb)( int ok, void *aux ) )
 {
@@ -520,77 +483,202 @@ socket_connect( conn_t *sock, void (*cb)( int ok, void *aux ) )
 		info( "\vok\n" );
 		socket_connected( sock );
 	} else {
-#ifdef HAVE_IPV6
-		int gaierr;
-		struct addrinfo hints;
-
-		memset( &hints, 0, sizeof(hints) );
-		hints.ai_family = AF_UNSPEC;
-		hints.ai_socktype = SOCK_STREAM;
-		hints.ai_flags = AI_ADDRCONFIG;
-		infon( "Resolving %s... ", conf->host );
-		if ((gaierr = getaddrinfo( conf->host, NULL, &hints, &sock->addrs ))) {
-			error( "Error: Cannot resolve server '%s': %s\n", conf->host, gai_strerror( gaierr ) );
-			socket_connect_bail( sock );
-			return;
-		}
-		info( "\vok\n" );
-#else
-		struct hostent *he;
-
-		infon( "Resolving %s... ", conf->host );
-		he = gethostbyname( conf->host );
-		if (!he) {
-			error( "Error: Cannot resolve server '%s': %s\n", conf->host, hstrerror( h_errno ) );
-			socket_connect_bail( sock );
-			return;
-		}
-		info( "\vok\n" );
-
-		sock->addrs = init_addrinfo( he );
-#endif
-		sock->curr_addr = sock->addrs;
-		socket_connect_one( sock );
+		socket_resolve( sock );
 	}
 }
 
+static void
+pipe_write( int fd, void *buf, int len )
+{
+	do {
+		int wrote = write( fd, buf, len );
+		if (wrote < 0) {
+			perror( "write" );
+			_exit( 1 );
+		}
+		buf = ((char *)buf) + wrote;
+		len -= wrote;
+	} while (len);
+}
+
+static void
+socket_resolve( conn_t *sock )
+{
+	info( "Resolving %s...\n", sock->conf->host );
+
+	int pfd[2];
+	if (pipe( pfd )) {
+		perror( "pipe" );
+		exit( 1 );
+	}
+	switch (fork()) {
+	case -1:
+		perror( "fork" );
+		exit( 1 );
+	case 0:
+		break;
+	default:
+		close( pfd[1] );
+		socket_register_internal( sock, pfd[0] );
+		sock->state = SCK_RESOLVING;
+		conf_notifier( &sock->notify, 0, POLLIN );
+		socket_expect_activity( sock, 1 );
+		return;
+	}
+
+#ifdef HAVE_IPV6
+	struct addrinfo *res, hints = { 0 };
+	hints.ai_family = AF_UNSPEC;
+	hints.ai_socktype = SOCK_STREAM;
+	hints.ai_flags = AI_ADDRCONFIG;
+	int gaierr = getaddrinfo( sock->conf->host, NULL, &hints, &res );
+	pipe_write( pfd[1], &gaierr, sizeof(gaierr) );
+	if (gaierr)
+		_exit( 1 );
+	static_assert( sizeof(((struct addrinfo){ 0 }).ai_family) == sizeof(int), "unexpected size of ai_family" );
+	static_assert( sizeof(struct in_addr) % sizeof(int) == 0, "unexpected size of struct in_addr" );
+	static_assert( sizeof(struct in6_addr) % sizeof(int) == 0, "unexpected size of struct in6_addr" );
+	int nbytes = 0;
+	for (struct addrinfo *cres = res; cres; cres = cres->ai_next) {
+		if (cres->ai_family == AF_INET) {
+			nbytes += sizeof(int) + sizeof(struct in_addr);
+		} else {
+			assert( cres->ai_family == AF_INET6 );
+			nbytes += sizeof(int) + sizeof(struct in6_addr);
+		}
+	}
+	pipe_write( pfd[1], &nbytes, sizeof(nbytes) );
+	for (struct addrinfo *cres = res; cres; cres = cres->ai_next) {
+		pipe_write( pfd[1], &cres->ai_family, sizeof(int) );
+		if (cres->ai_family == AF_INET)
+			pipe_write( pfd[1], &((struct sockaddr_in *)cres->ai_addr)->sin_addr, sizeof(struct in_addr) );
+		else
+			pipe_write( pfd[1], &((struct sockaddr_in6 *)cres->ai_addr)->sin6_addr, sizeof(struct in6_addr) );
+	}
+#else
+	struct hostent *he = gethostbyname( sock->conf->host );
+	int herrno = he ? 0 : h_errno;
+	pipe_write( pfd[1], &herrno, sizeof(herrno) );
+	if (!he)
+		_exit( 1 );
+	static_assert( sizeof(struct in_addr) % sizeof(int) == 0, "unexpected size of struct in_addr" );
+	int nbytes = 0;
+	for (char **addr = he->h_addr_list; *addr; addr++)
+		nbytes += sizeof(struct in_addr);
+	pipe_write( pfd[1], &nbytes, sizeof(nbytes) );
+	for (char **addr = he->h_addr_list; *addr; addr++)
+		pipe_write( pfd[1], *addr, sizeof(struct in_addr) );
+#endif
+	_exit( 0 );
+}
+
+static void
+pipe_read( int fd, void *buf, int len )
+{
+	do {
+		int didrd = read( fd, buf, len );
+		if (didrd < 0) {
+			sys_error( "read" );
+			exit( 1 );
+		}
+		if (!didrd) {
+			error( "read: unexpected EOF\n" );
+			exit( 1 );
+		}
+		buf = ((char *)buf) + didrd;
+		len -= didrd;
+	} while (len);
+}
+
+static void
+socket_resolve_finalize( conn_t *sock )
+{
+	int errcode;
+	pipe_read( sock->fd, &errcode, sizeof(errcode) );
+	if (errcode) {
+#ifdef HAVE_IPV6
+		const char *err = gai_strerror( errcode );
+#else
+		const char *err = hstrerror( errcode );
+#endif
+		error( "Error: Cannot resolve server '%s': %s\n", sock->conf->host, err );
+		socket_close_internal( sock );
+		socket_connect_bail( sock );
+		return;
+	}
+
+	int nbytes;
+	pipe_read( sock->fd, &nbytes, sizeof(nbytes) );
+	char *addrs = nfmalloc( nbytes );
+	pipe_read( sock->fd, addrs, nbytes );
+	sock->curr_addr = sock->addrs = addrs;
+	sock->addrs_end = addrs + nbytes;
+	socket_close_internal( sock );  // Get rid of the pipe
+	socket_connect_one( sock );
+}
+
+static void
+socket_resolve_timeout( conn_t *sock )
+{
+	error( "Error: Cannot resolve server '%s': timeout.\n", sock->conf->host );
+	socket_close_internal( sock );
+	socket_connect_bail( sock );
+}
+
 static void
 socket_connect_one( conn_t *sock )
 {
-	int s;
-#ifdef HAVE_IPV6
-	struct addrinfo *ai;
-#else
-	struct addr_info *ai;
-#endif
-
-	if (!(ai = sock->curr_addr)) {
+	char *ai = sock->curr_addr;
+	if (ai == sock->addrs_end) {
 		error( "No working address found for %s\n", sock->conf->host );
 		socket_connect_bail( sock );
 		return;
 	}
 
+	union {
+		struct sockaddr any;
+		struct sockaddr_in ip4;
 #ifdef HAVE_IPV6
-	if (ai->ai_family == AF_INET6) {
-		struct sockaddr_in6 *in6 = ((struct sockaddr_in6 *)ai->ai_addr);
+		struct sockaddr_in6 ip6;
+#endif
+	} addr;
+
+#ifdef HAVE_IPV6
+	int fam = *(int *)ai;
+	ai += sizeof(int);
+	int addr_len;
+	if (fam == AF_INET6) {
+		addr_len = sizeof(addr.ip6);
+		addr.ip6.sin6_addr = *(struct in6_addr *)ai;
+		addr.ip6.sin6_flowinfo = 0;
+		addr.ip6.sin6_scope_id = 0;
+		ai += sizeof(struct in6_addr);
+	} else {
+		addr_len = sizeof(addr.ip4);
+#else
+	const int fam = AF_INET;
+	const int addr_len = sizeof(addr.ip4);
+	{
+#endif
+		addr.ip4.sin_addr = *(struct in_addr *)ai;
+		ai += sizeof(struct in_addr);
+	}
+	sock->curr_addr = ai;
+
+#ifdef HAVE_IPV6
+	if (fam == AF_INET6) {
 		char sockname[64];
-		in6->sin6_port = htons( sock->conf->port );
+		inet_ntop( fam, &addr.ip6.sin6_addr, sockname, sizeof(sockname) );
 		nfasprintf( &sock->name, "%s ([%s]:%hu)",
-		            sock->conf->host, inet_ntop( AF_INET6, &in6->sin6_addr, sockname, sizeof(sockname) ), sock->conf->port );
+		            sock->conf->host, sockname, sock->conf->port );
 	} else
 #endif
 	{
-		struct sockaddr_in *in = ((struct sockaddr_in *)ai->ai_addr);
-		in->sin_port = htons( sock->conf->port );
 		nfasprintf( &sock->name, "%s (%s:%hu)",
-		            sock->conf->host, inet_ntoa( in->sin_addr ), sock->conf->port );
+		            sock->conf->host, inet_ntoa( addr.ip4.sin_addr ), sock->conf->port );
 	}
 
-#ifdef HAVE_IPV6
-	s = socket( ai->ai_family, SOCK_STREAM, 0 );
-#else
-	s = socket( PF_INET, SOCK_STREAM, 0 );
-#endif
+	int s = socket( fam, SOCK_STREAM, 0 );
 	if (s < 0) {
 		socket_connect_next( sock );
 		return;
@@ -598,11 +686,9 @@ socket_connect_one( conn_t *sock )
 	socket_open_internal( sock, s );
 
 	infon( "Connecting to %s... ", sock->name );
-#ifdef HAVE_IPV6
-	if (connect( s, ai->ai_addr, ai->ai_addrlen )) {
-#else
-	if (connect( s, ai->ai_addr, sizeof(*ai->ai_addr) )) {
-#endif
+	addr.any.sa_family = fam;
+	addr.ip4.sin_port = htons( sock->conf->port );  // Aliased for ip6
+	if (connect( s, &addr.any, addr_len )) {
 		if (errno != EINPROGRESS) {
 			socket_connect_failed( sock );
 			return;
@@ -623,7 +709,6 @@ socket_connect_next( conn_t *conn )
 	sys_error( "Cannot connect to %s", conn->name );
 	free( conn->name );
 	conn->name = NULL;
-	conn->curr_addr = conn->curr_addr->ai_next;
 	socket_connect_one( conn );
 }
 
@@ -637,10 +722,8 @@ socket_connect_failed( conn_t *conn )
 static void
 socket_connected( conn_t *conn )
 {
-	if (conn->addrs) {
-		freeaddrinfo( conn->addrs );
-		conn->addrs = NULL;
-	}
+	free( conn->addrs );
+	conn->addrs = NULL;
 	conf_notifier( &conn->notify, 0, POLLIN );
 	socket_expect_activity( conn, 0 );
 	conn->state = SCK_READY;
@@ -650,10 +733,8 @@ socket_connected( conn_t *conn )
 static void
 socket_cleanup_names( conn_t *conn )
 {
-	if (conn->addrs) {
-		freeaddrinfo( conn->addrs );
-		conn->addrs = NULL;
-	}
+	free( conn->addrs );
+	conn->addrs = NULL;
 	free( conn->name );
 	conn->name = NULL;
 }
@@ -739,6 +820,34 @@ do_read( conn_t *sock, char *buf, uint len )
 	return n;
 }
 
+static void
+socket_filled( conn_t *conn, uint len )
+{
+	uint off = conn->offset;
+	uint cnt = conn->bytes + len;
+	conn->bytes = cnt;
+	if (conn->wanted) {
+		// Fulfill as much of the request as still fits into the buffer,
+		// but avoid chopping up the actual socket reads
+		if (cnt < conn->wanted && off + cnt < sizeof(conn->buf) - conn->readsz)
+			return;
+	} else {
+		// Need a full line
+		char *s = conn->buf + off;
+		char *p = memchr( s + conn->scanoff, '\n', cnt - conn->scanoff );
+		if (!p) {
+			conn->scanoff = cnt;
+			if (off && off + cnt >= sizeof(conn->buf) - conn->readsz) {
+				memmove( conn->buf, conn->buf + off, cnt );
+				conn->offset = 0;
+			}
+			return;
+		}
+		conn->scanoff = (uint)(p - s);
+	}
+	conn->read_callback( conn->callback_aux );
+}
+
 #ifdef HAVE_LIBZ
 static void
 socket_fill_z( conn_t *sock )
@@ -765,10 +874,8 @@ socket_fill_z( conn_t *sock )
 	if (!sock->in_z->avail_out)
 		conf_wakeup( &sock->z_fake, 0 );
 
-	if ((len = (uint)((char *)sock->in_z->next_out - buf))) {
-		sock->bytes += len;
-		sock->read_callback( sock->callback_aux );
-	}
+	if ((len = (uint)((char *)sock->in_z->next_out - buf)))
+		socket_filled( sock, len );
 }
 #endif
 
@@ -798,8 +905,16 @@ socket_fill( conn_t *sock )
 		if ((n = do_read( sock, buf, len )) <= 0)
 			return;
 
-		sock->bytes += (uint)n;
-		sock->read_callback( sock->callback_aux );
+		// IIR filter for tracking average size of bulk reads.
+		// We use this to optimize the free space at the end of the
+		// buffer, hence the factor of 1.5.
+		if (n >= MIN_BULK_READ) {
+			sock->readsz = (sock->readsz * 3 + n * 3 / 2) / 4;
+			if (sock->readsz > sizeof(sock->buf))
+				sock->readsz = sizeof(sock->buf);
+		}
+
+		socket_filled( sock, (uint)n );
 	}
 }
 
@@ -810,44 +925,70 @@ socket_expect_activity( conn_t *conn, int expect )
 		conf_wakeup( &conn->fd_timeout, expect ? conn->conf->timeout : -1 );
 }
 
-int
-socket_read( conn_t *conn, char *buf, uint len )
+void
+socket_expect_eof( conn_t *sock )
 {
-	uint n = conn->bytes;
-	if (!n && conn->state == SCK_EOF)
-		return -1;
-	if (n > len)
-		n = len;
-	memcpy( buf, conn->buf + conn->offset, n );
-	if (!(conn->bytes -= n))
-		conn->offset = 0;
-	else
-		conn->offset += n;
-	return (int)n;
+#ifdef SSL_OP_IGNORE_UNEXPECTED_EOF  // implies HAVE_LIBSSL
+	if (sock->ssl)
+		SSL_set_options( sock->ssl, SSL_OP_IGNORE_UNEXPECTED_EOF );
+#endif
+}
+
+void
+socket_expect_bytes( conn_t *conn, uint len )
+{
+	conn->wanted = len;
+	uint off = conn->offset;
+	if (off) {
+		uint cnt = conn->bytes;
+		if (off + len > sizeof(conn->buf) ||
+		    off + cnt >= sizeof(conn->buf) - conn->readsz) {
+			memmove( conn->buf, conn->buf + off, cnt );
+			conn->offset = 0;
+		}
+	}
 }
 
 char *
-socket_read_line( conn_t *b )
+socket_read( conn_t *conn, uint min_len, uint max_len, uint *out_len )
 {
-	char *p, *s;
-	uint n;
+	assert( min_len > 0 );
+	assert( min_len <= sizeof(conn->buf) );
+	assert( min_len <= max_len );
 
-	s = b->buf + b->offset;
-	p = memchr( s + b->scanoff, '\n', b->bytes - b->scanoff );
-	if (!p) {
-		b->scanoff = b->bytes;
-		if (b->offset + b->bytes == sizeof(b->buf)) {
-			memmove( b->buf, b->buf + b->offset, b->bytes );
-			b->offset = 0;
-		}
-		if (b->state == SCK_EOF)
+	uint off = conn->offset;
+	uint cnt = conn->bytes;
+	if (cnt < min_len) {
+		if (conn->state == SCK_EOF)
 			return (void *)~0;
 		return NULL;
 	}
-	n = (uint)(p + 1 - s);
-	b->offset += n;
-	b->bytes -= n;
-	b->scanoff = 0;
+	uint n = (cnt < max_len) ? cnt : max_len;
+	cnt -= n;
+	conn->offset = cnt ? off + n : 0;
+	conn->bytes = cnt;
+	*out_len = n;
+	return conn->buf + off;
+}
+
+char *
+socket_read_line( conn_t *conn )
+{
+	uint off = conn->offset;
+	uint cnt = conn->bytes;
+	char *s = conn->buf + off;
+	char *p = memchr( s + conn->scanoff, '\n', cnt - conn->scanoff );
+	if (!p) {
+		if (conn->state == SCK_EOF)
+			return (void *)~0;
+		conn->scanoff = cnt;
+		return NULL;
+	}
+	uint n = (uint)(p + 1 - s);
+	cnt -= n;
+	conn->offset = cnt ? off + n : 0;
+	conn->bytes = cnt;
+	conn->scanoff = 0;
 	if (p != s && p[-1] == '\r')
 		p--;
 	*p = 0;
@@ -1068,6 +1209,11 @@ socket_fd_cb( int events, void *aux )
 {
 	conn_t *conn = (conn_t *)aux;
 
+	if (conn->state == SCK_RESOLVING) {
+		socket_resolve_finalize( conn );
+		return;
+	}
+
 	if ((events & POLLERR) || conn->state == SCK_CONNECTING) {
 		int soerr;
 		socklen_t selen = sizeof(soerr);
@@ -1130,7 +1276,9 @@ socket_timeout_cb( void *aux )
 {
 	conn_t *conn = (conn_t *)aux;
 
-	if (conn->state == SCK_CONNECTING) {
+	if (conn->state == SCK_RESOLVING) {
+		socket_resolve_timeout( conn );
+	} else if (conn->state == SCK_CONNECTING) {
 		errno = ETIMEDOUT;
 		socket_connect_failed( conn );
 	} else {
