@@ -1,23 +1,8 @@
+// SPDX-FileCopyrightText: 2000-2002 Michael R. Elkins <me@mutt.org>
+// SPDX-FileCopyrightText: 2002-2022 Oswald Buddenhagen <ossi@users.sf.net>
+// SPDX-License-Identifier: GPL-2.0-or-later WITH LicenseRef-isync-GPL-exception
 /*
  * mbsync - mailbox synchronizer
- * Copyright (C) 2000-2002 Michael R. Elkins <me@mutt.org>
- * Copyright (C) 2002-2006,2010-2012 Oswald Buddenhagen <ossi@users.sf.net>
- *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
- * As a special exception, mbsync may be linked with the OpenSSL library,
- * despite that library's more restrictive license.
  */
 
 #ifndef SOCKET_H
@@ -72,11 +57,7 @@ typedef struct {
 	int fd;
 	int state;
 	const server_conf_t *conf; /* needed during connect */
-#ifdef HAVE_IPV6
-	struct addrinfo *addrs, *curr_addr; /* needed during connect */
-#else
-	struct addr_info *addrs, *curr_addr; /* needed during connect */
-#endif
+	char *addrs, *addrs_end, *curr_addr;  // needed during connect; assumed to be int-aligned
 	char *name;
 #ifdef HAVE_LIBSSL
 	SSL *ssl;
@@ -114,12 +95,17 @@ typedef struct {
 	uint offset; /* start of filled bytes in buffer */
 	uint bytes; /* number of filled bytes in buffer */
 	uint scanoff; /* offset to continue scanning for newline at, relative to 'offset' */
+	uint wanted;  // try to accumulate that many bytes before calling back; 0 => full line
+	uint readsz;  // average size of bulk reads from the underlying socket, times 1.5
 	char buf[100000];
 #ifdef HAVE_LIBZ
 	char z_buf[100000];
 #endif
 } conn_t;
 
+// Shorter reads are assumed to be limited by round-trips.
+#define MIN_BULK_READ 1000
+
 /* call this before doing anything with the socket */
 static INLINE void socket_init( conn_t *conn,
                                 const server_conf_t *conf,
@@ -136,14 +122,19 @@ static INLINE void socket_init( conn_t *conn,
 	conn->fd = -1;
 	conn->name = NULL;
 	conn->write_buf_append = &conn->write_buf;
+	conn->wanted = 1;
+	conn->readsz = MIN_BULK_READ * 3 / 2;
 }
 void socket_connect( conn_t *conn, void (*cb)( int ok, void *aux ) );
 void socket_start_tls(conn_t *conn, void (*cb)( int ok, void *aux ) );
 void socket_start_deflate( conn_t *conn );
 void socket_close( conn_t *sock );
 void socket_expect_activity( conn_t *sock, int expect );
-int socket_read( conn_t *sock, char *buf, uint len ); /* never waits */
-char *socket_read_line( conn_t *sock ); /* don't free return value; never waits */
+void socket_expect_eof( conn_t *sock );
+void socket_expect_bytes( conn_t *sock, uint len );
+// Don't free return values. These functions never wait.
+char *socket_read( conn_t *conn, uint min_len, uint max_len, uint *out_len );
+char *socket_read_line( conn_t *conn );
 typedef enum { KeepOwn = 0, GiveOwn } ownership_t;
 typedef struct {
 	char *buf;
