@@ -1,38 +1,29 @@
+// SPDX-FileCopyrightText: 2000-2002 Michael R. Elkins <me@mutt.org>
+// SPDX-FileCopyrightText: 2002-2022 Oswald Buddenhagen <ossi@users.sf.net>
+// SPDX-License-Identifier: GPL-2.0-or-later WITH LicenseRef-isync-GPL-exception
 /*
  * mbsync - mailbox synchronizer
- * Copyright (C) 2000-2002 Michael R. Elkins <me@mutt.org>
- * Copyright (C) 2002-2006,2011,2012 Oswald Buddenhagen <ossi@users.sf.net>
- *
- *  This program is free software; you can redistribute it and/or modify
- *  it under the terms of the GNU General Public License as published by
- *  the Free Software Foundation; either version 2 of the License, or
- *  (at your option) any later version.
- *
- *  This program is distributed in the hope that it will be useful,
- *  but WITHOUT ANY WARRANTY; without even the implied warranty of
- *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU General Public License for more details.
- *
- *  You should have received a copy of the GNU General Public License
- *  along with this program.  If not, see <http://www.gnu.org/licenses/>.
- *
- * As a special exception, mbsync may be linked with the OpenSSL library,
- * despite that library's more restrictive license.
  */
 
 #include "common.h"
 
-#include <assert.h>
-#include <stddef.h>
-#include <stdlib.h>
-#include <unistd.h>
 #include <fcntl.h>
+#include <sys/stat.h>
 #include <errno.h>
-#include <string.h>
 #include <ctype.h>
 #include <pwd.h>
+#include <sys/time.h>
 
-static int need_nl;
+int Verbosity = TERSE;
+int DFlags;
+int JLimit, JCount;
+int UseFSync = 1;
+
+int Pid;
+char Hostname[256];
+const char *Home;
+
+static int need_nl, need_del;
 
 void
 flushn( void )
@@ -41,38 +32,52 @@ flushn( void )
 		putchar( '\n' );
 		fflush( stdout );
 		need_nl = 0;
+	} else if (need_del) {
+		static const char delstr[] =
+		        "                                                            "
+		        "                                                            ";
+		if (need_del > (int)sizeof(delstr) - 1)
+			need_del = (int)sizeof(delstr) - 1;
+		// We could use ^[[K instead, but we assume a dumb terminal.
+		printf( "\r%.*s\r", need_del, delstr );
+		fflush( stdout );
+		need_del = 0;
 	}
 }
 
 static void ATTR_PRINTFLIKE(1, 0)
-printn( const char *msg, va_list va )
+vprint( const char *msg, va_list va )
 {
-	if (*msg == '\v')
-		msg++;
-	else
-		flushn();
 	vprintf( msg, va );
 	fflush( stdout );
+	need_nl = 0;
 }
 
 void
-vdebug( int cat, const char *msg, va_list va )
+print( const char *msg, ... )
 {
-	if (DFlags & cat) {
-		vprintf( msg, va );
-		fflush( stdout );
-		need_nl = 0;
-	}
+	va_list va;
+
+	va_start( va, msg );
+	vprint( msg, va );
+	va_end( va );
+}
+
+static void ATTR_PRINTFLIKE(1, 0)
+vprintn( const char *msg, va_list va )
+{
+	vprint( msg, va );
+	need_nl = 1;
 }
 
 void
-vdebugn( int cat, const char *msg, va_list va )
+printn( const char *msg, ... )
 {
-	if (DFlags & cat) {
-		vprintf( msg, va );
-		fflush( stdout );
-		need_nl = 1;
-	}
+	va_list va;
+
+	va_start( va, msg );
+	vprintn( msg, va );
+	va_end( va );
 }
 
 void
@@ -81,10 +86,19 @@ progress( const char *msg, ... )
 	va_list va;
 
 	va_start( va, msg );
-	vprintf( msg, va );
+	need_del = vprintf( msg, va ) - 1;
 	va_end( va );
 	fflush( stdout );
-	need_nl = 1;
+}
+
+static void ATTR_PRINTFLIKE(1, 0)
+nvprint( const char *msg, va_list va )
+{
+	if (*msg == '\v')
+		msg++;
+	else
+		flushn();
+	vprint( msg, va );
 }
 
 void
@@ -92,11 +106,10 @@ info( const char *msg, ... )
 {
 	va_list va;
 
-	if (DFlags & VERBOSE) {
+	if (Verbosity >= VERBOSE) {
 		va_start( va, msg );
-		printn( msg, va );
+		nvprint( msg, va );
 		va_end( va );
-		need_nl = 0;
 	}
 }
 
@@ -105,9 +118,9 @@ infon( const char *msg, ... )
 {
 	va_list va;
 
-	if (DFlags & VERBOSE) {
+	if (Verbosity >= VERBOSE) {
 		va_start( va, msg );
-		printn( msg, va );
+		nvprint( msg, va );
 		va_end( va );
 		need_nl = 1;
 	}
@@ -118,11 +131,10 @@ notice( const char *msg, ... )
 {
 	va_list va;
 
-	if (!(DFlags & QUIET)) {
+	if (Verbosity >= TERSE) {
 		va_start( va, msg );
-		printn( msg, va );
+		nvprint( msg, va );
 		va_end( va );
-		need_nl = 0;
 	}
 }
 
@@ -131,7 +143,7 @@ warn( const char *msg, ... )
 {
 	va_list va;
 
-	if (!(DFlags & VERYQUIET)) {
+	if (Verbosity >= QUIET) {
 		flushn();
 		va_start( va, msg );
 		vfprintf( stderr, msg, va );
@@ -173,6 +185,225 @@ sys_error( const char *msg, ... )
 	va_end( va );
 }
 
+// Minimal printf() replacement with custom format sequence(s):
+// - %\\s
+//   Print backslash-escaped string literals. Note that this does not
+//   automatically add quotes around the printed string, so it is
+//   possible to concatenate multiple segments.
+// - %!s
+//   Same as %\\s, but non-ASCII characters are (hex-)escaped as well.
+// - %!&s
+//   Same as %!s, but linefeeds are also printed verbatim for legibility.
+
+// TODO: Trade off segments vs. buffer capacity dynamically.
+#define QPRINTF_SEGS 16
+#ifndef QPRINTF_BUFF
+# define QPRINTF_BUFF 1000
+#endif
+
+typedef void (*printf_cb)( const char **segs, uint *segls, int nsegs, uint totlen, void *aux );
+
+static void
+xvprintf_core( const char *fmt, va_list ap, printf_cb cb, void *cb_aux )
+{
+	int nsegs = 0;
+	uint totlen = 0;
+	const char *segs[QPRINTF_SEGS];
+	uint segls[QPRINTF_SEGS];
+	char buf[QPRINTF_BUFF];
+
+#define ADD_SEG(p, l) \
+		do { \
+			if (nsegs == QPRINTF_SEGS) \
+				oob(); \
+			segs[nsegs] = p; \
+			segls[nsegs++] = l; \
+			totlen += l; \
+		} while (0)
+
+	char *d = buf;
+	char *ed = d + sizeof(buf);
+	const char *s = fmt;
+	for (;;) {
+		char c = *fmt;
+		if (!c || c == '%') {
+			uint l = fmt - s;
+			if (l)
+				ADD_SEG( s, l );
+			if (!c)
+				break;
+			uint maxlen = UINT_MAX;
+			c = *++fmt;
+			if (c == '.') {
+				c = *++fmt;
+				if (c != '*') {
+					fputs( "Fatal: unsupported string length specification. Please report a bug.\n", stderr );
+					abort();
+				}
+				maxlen = va_arg( ap, uint );
+				c = *++fmt;
+			}
+			int escaped = 0;
+			if (c == '\\') {
+				escaped = 1;
+				c = *++fmt;
+			} else if (c == '!') {
+				escaped = 2;
+				c = *++fmt;
+				if (c == '&') {
+					escaped = 3;
+					c = *++fmt;
+				}
+			}
+			if (c == 'c') {
+				if (d + 1 > ed)
+					oob();
+				ADD_SEG( d, 1 );
+				*d++ = (char)va_arg( ap, int );
+			} else if (c == 's') {
+				s = va_arg( ap, const char * );
+				if (escaped) {
+					char *bd = d;
+					for (l = 0; l < maxlen && (c = *s); l++, s++) {
+						if (c == '\\' || c == '"') {
+							if (d >= ed)
+								oob();
+							*d++ = '\\';
+						} else if (escaped >= 2 && (c < 32 || c > 126)) {
+							switch (c) {
+							case '\r': c = 'r'; break;
+							case '\t': c = 't'; break;
+							case '\a': c = 'a'; break;
+							case '\b': c = 'b'; break;
+							case '\v': c = 'v'; break;
+							case '\f': c = 'f'; break;
+							case '\n':
+								if (escaped == 2) {
+									c = 'n';
+									break;
+								}
+								if (d + 2 >= ed)
+									oob();
+								*d++ = '\\';
+								*d++ = 'n';
+								*d++ = c;  // Keep the actual line break for legibility.
+								continue;
+							default:
+								d += nfsnprintf( d, ed - d, "\\x%02x", (uchar)c );
+								continue;
+							}
+							if (d >= ed)
+								oob();
+							*d++ = '\\';
+						}
+						if (d >= ed)
+							oob();
+						*d++ = c;
+					}
+					l = d - bd;
+					if (l)
+						ADD_SEG( bd, l );
+				} else {
+					l = strnlen( s, maxlen );
+					if (l)
+						ADD_SEG( s, l );
+				}
+			} else if (c == 'd') {
+				l = nfsnprintf( d, ed - d, "%d", va_arg( ap, int ) );
+				ADD_SEG( d, l );
+				d += l;
+			} else if (c == 'u') {
+				l = nfsnprintf( d, ed - d, "%u", va_arg( ap, uint ) );
+				ADD_SEG( d, l );
+				d += l;
+			} else {
+				fputs( "Fatal: unsupported format specifier. Please report a bug.\n", stderr );
+				abort();
+			}
+			s = ++fmt;
+		} else {
+			fmt++;
+		}
+	}
+	cb( segs, segls, nsegs, totlen, cb_aux );
+}
+
+static void
+xasprintf_cb( const char **segs, uint *segls, int nsegs, uint totlen, void *aux )
+{
+	char *d = nfmalloc( totlen + 1 );
+	*(char **)aux = d;
+	for (int i = 0; i < nsegs; i++) {
+		memcpy( d, segs[i], segls[i] );
+		d += segls[i];
+	}
+	*d = 0;
+}
+
+char *
+xvasprintf( const char *fmt, va_list ap )
+{
+	char *out;
+	xvprintf_core( fmt, ap, xasprintf_cb, &out );
+	return out;
+}
+
+#ifndef HAVE_FWRITE_UNLOCKED
+# define flockfile(f)
+# define funlockfile(f)
+# define fwrite_unlocked(b, l, n, f) fwrite(b, l, n, f)
+#endif
+
+static void
+xprintf_cb( const char **segs, uint *segls, int nsegs, uint totlen ATTR_UNUSED, void *aux ATTR_UNUSED )
+{
+	flockfile( stdout );
+	for (int i = 0; i < nsegs; i++)
+		fwrite_unlocked( segs[i], 1, segls[i], stdout );
+	funlockfile( stdout );
+}
+
+void
+xprintf( const char *fmt, ... )
+{
+	va_list va;
+
+	va_start( va, fmt );
+	xvprintf_core( fmt, va, xprintf_cb, NULL );
+	va_end( va );
+}
+
+void
+vFprintf( FILE *f, const char *msg, va_list va )
+{
+	int r;
+
+	r = vfprintf( f, msg, va );
+	if (r < 0) {
+		sys_error( "Error: cannot write file" );
+		exit( 1 );
+	}
+}
+
+void
+Fprintf( FILE *f, const char *msg, ... )
+{
+	va_list va;
+
+	va_start( va, msg );
+	vFprintf( f, msg, va );
+	va_end( va );
+}
+
+void
+Fclose( FILE *f, int safe )
+{
+	if ((safe && (fflush( f ) || (UseFSync && fdatasync( fileno( f ) )))) || fclose( f ) == EOF) {
+		sys_error( "Error: cannot close file" );
+		exit( 1 );
+	}
+}
+
 void
 add_string_list_n( string_list_t **list, const char *str, uint len )
 {
@@ -242,6 +473,13 @@ strnlen( const char *str, size_t maxlen )
 
 #endif
 
+void
+to_upper( char *str, uint len )
+{
+	for (uint i = 0; i < len; i++)
+		str[i] = toupper( str[i] );
+}
+
 int
 starts_with( const char *str, int strl, const char *cmp, uint cmpl )
 {
@@ -250,6 +488,15 @@ starts_with( const char *str, int strl, const char *cmp, uint cmpl )
 	return ((uint)strl >= cmpl) && !memcmp( str, cmp, cmpl );
 }
 
+static int
+equals_upper_impl( const char *str, const char *cmp, uint cmpl )
+{
+	for (uint i = 0; i < cmpl; i++)
+		if (toupper( str[i] ) != cmp[i])
+			return 0;
+	return 1;
+}
+
 int
 starts_with_upper( const char *str, int strl, const char *cmp, uint cmpl )
 {
@@ -257,10 +504,7 @@ starts_with_upper( const char *str, int strl, const char *cmp, uint cmpl )
 		strl = strnlen( str, cmpl + 1 );
 	if ((uint)strl < cmpl)
 		return 0;
-	for (uint i = 0; i < cmpl; i++)
-		if (str[i] != cmp[i] && toupper( str[i] ) != cmp[i])
-			return 0;
-	return 1;
+	return equals_upper_impl( str, cmp, cmpl );
 }
 
 int
@@ -271,6 +515,16 @@ equals( const char *str, int strl, const char *cmp, uint cmpl )
 	return ((uint)strl == cmpl) && !memcmp( str, cmp, cmpl );
 }
 
+int
+equals_upper( const char *str, int strl, const char *cmp, uint cmpl )
+{
+	if (strl < 0)
+		strl = strnlen( str, cmpl + 1 );
+	if ((uint)strl != cmpl)
+		return 0;
+	return equals_upper_impl( str, cmp, cmpl );
+}
+
 #ifndef HAVE_TIMEGM
 /*
    Converts struct tm to time_t, assuming the data in tm is UTC rather
@@ -336,6 +590,21 @@ timegm( struct tm *t )
 }
 #endif
 
+void
+fmt_bits( uint bits, uint num_bits, const char *bit_str, const int *bit_off, char *buf )
+{
+	uint d = 0;
+	for (uint i = 0, val = 1; i < num_bits; i++, val <<= 1) {
+		if (bits & val) {
+			if (d)
+				buf[d++] = ',';
+			for (const char *s = bit_str + bit_off[i]; *s; s++)
+				buf[d++] = *s;
+		}
+	}
+	buf[d] = 0;
+}
+
 void
 oob( void )
 {
@@ -374,7 +643,7 @@ nfmalloc( size_t sz )
 }
 
 void *
-nfcalloc( size_t sz )
+nfzalloc( size_t sz )
 {
 	void *ret;
 
@@ -449,53 +718,23 @@ cur_user( void )
 }
 */
 
-char *
-expand_strdup( const char *s )
-{
-	struct passwd *pw;
-	const char *p, *q;
-	char *r;
-
-	if (*s == '~') {
-		s++;
-		if (!*s) {
-			p = NULL;
-			q = Home;
-		} else if (*s == '/') {
-			p = s;
-			q = Home;
-		} else {
-			if ((p = strchr( s, '/' ))) {
-				r = nfstrndup( s, (size_t)(p - s) );
-				pw = getpwnam( r );
-				free( r );
-			} else
-				pw = getpwnam( s );
-			if (!pw)
-				return NULL;
-			q = pw->pw_dir;
-		}
-		nfasprintf( &r, "%s%s", q, p ? p : "" );
-		return r;
-	} else
-		return nfstrdup( s );
-}
-
 /* Return value: 0 = ok, -1 = out found in arg, -2 = in found in arg but no out specified */
 int
-map_name( const char *arg, char **result, uint reserve, const char *in, const char *out )
+map_name( const char *arg, int l, char **result, uint reserve, const char *in, const char *out )
 {
 	char *p;
-	uint i, l, ll, num, inl, outl;
+	int i, ll, num, inl, outl;
 
 	assert( arg );
-	l = strlen( arg );
+	if (l < 0)
+		l = strlen( arg );
 	assert( in );
 	inl = strlen( in );
 	if (!inl) {
 	  copy:
 		*result = nfmalloc( reserve + l + 1 );
-		memcpy( *result + reserve, arg, l + 1 );
+		memcpy( *result + reserve, arg, l );
+		(*result)[reserve + l] = 0;
 		return 0;
 	}
 	assert( out );
@@ -503,6 +742,8 @@ map_name( const char *arg, char **result, uint reserve, const char *in, const ch
 	if (equals( in, (int)inl, out, outl ))
 		goto copy;
 	for (num = 0, i = 0; i < l; ) {
+		if (i + inl > l)
+			goto fout;
 		for (ll = 0; ll < inl; ll++)
 			if (arg[i + ll] != in[ll])
 				goto fout;
@@ -511,6 +752,8 @@ map_name( const char *arg, char **result, uint reserve, const char *in, const ch
 		continue;
 	  fout:
 		if (outl) {
+			if (i + outl > l)
+				goto fnexti;
 			for (ll = 0; ll < outl; ll++)
 				if (arg[i + ll] != out[ll])
 					goto fnexti;
@@ -526,6 +769,8 @@ map_name( const char *arg, char **result, uint reserve, const char *in, const ch
 	*result = nfmalloc( reserve + l + num * (outl - inl) + 1 );
 	p = *result + reserve;
 	for (i = 0; i < l; ) {
+		if (i + inl > l)
+			goto rnexti;
 		for (ll = 0; ll < inl; ll++)
 			if (arg[i + ll] != in[ll])
 				goto rnexti;
@@ -540,6 +785,21 @@ map_name( const char *arg, char **result, uint reserve, const char *in, const ch
 	return 0;
 }
 
+int
+mkdir_p( char *path, int len )
+{
+	if (!mkdir( path, 0700 ) || errno == EEXIST)
+		return 0;
+	char *p = memrchr( path, '/', (size_t)len );
+	*p = 0;
+	if (mkdir_p( path, (int)(p - path) )) {
+		*p = '/';
+		return -1;
+	}
+	*p = '/';
+	return mkdir( path, 0700 );
+}
+
 static int
 compare_uints( const void *l, const void *r )
 {
@@ -741,10 +1001,41 @@ wipe_notifier( notifier_t *sn )
 #endif
 }
 
-static time_t
+#if _POSIX_TIMERS - 0 > 0
+static clockid_t clkid;
+#endif
+
+void
+init_timers( void )
+{
+#if _POSIX_TIMERS - 0 > 0
+	struct timespec ts;
+# ifdef CLOCK_BOOTTIME
+	if (!clock_gettime( CLOCK_BOOTTIME, &ts )) {
+		clkid = CLOCK_BOOTTIME;
+	} else
+# endif
+# ifdef CLOCK_MONOTONIC_COARSE
+	if (!clock_gettime( CLOCK_MONOTONIC_COARSE, &ts )) {
+		clkid = CLOCK_MONOTONIC_COARSE;
+	} else
+# endif
+	clkid = CLOCK_MONOTONIC;
+#endif
+}
+
+int64_t
 get_now( void )
 {
-	return time( NULL );
+#if _POSIX_TIMERS - 0 > 0
+	struct timespec ts;
+	clock_gettime( clkid, &ts );
+	return ts.tv_sec * 1000LL + ts.tv_nsec / 1000000;
+#else
+	struct timeval tv;
+	gettimeofday( &tv, NULL );
+	return tv.tv_sec * 1000LL + tv.tv_usec / 1000;
+#endif
 }
 
 static list_head_t timers = { &timers, &timers };
@@ -773,7 +1064,7 @@ conf_wakeup( wakeup_t *tmr, int to )
 		if (tmr->links.next)
 			list_unlink( &tmr->links );
 	} else {
-		time_t timeout = to;
+		int64_t timeout = to;
 		if (!to) {
 			/* We always prepend null timers, to cluster related events. */
 			succ = timers.next;
@@ -807,18 +1098,20 @@ event_wait( void )
 	int timeout = -1;
 	if ((head = timers.next) != &timers) {
 		wakeup_t *tmr = (wakeup_t *)head;
-		time_t delta = tmr->timeout;
+		int64_t delta = tmr->timeout;
 		if (!delta || (delta -= get_now()) <= 0) {
 			list_unlink( head );
 			tmr->cb( tmr->aux );
 			return;
 		}
-		timeout = (int)delta * 1000;
+		timeout = (int)delta;
 	}
 	switch (poll( pollfds, npolls, timeout )) {
 	case 0:
 		return;
 	case -1:
+		if (errno == EINTR)
+			return;
 		perror( "poll() failed in event loop" );
 		abort();
 	default:
@@ -843,14 +1136,14 @@ event_wait( void )
 
 	if ((head = timers.next) != &timers) {
 		wakeup_t *tmr = (wakeup_t *)head;
-		time_t delta = tmr->timeout;
+		int64_t delta = tmr->timeout;
 		if (!delta || (delta -= get_now()) <= 0) {
 			list_unlink( head );
 			tmr->cb( tmr->aux );
 			return;
 		}
-		to_tv.tv_sec = delta;
-		to_tv.tv_usec = 0;
+		to_tv.tv_sec = delta / 1000;
+		to_tv.tv_usec = delta * 1000;
 		timeout = &to_tv;
 	}
 	FD_ZERO( &rfds );
@@ -871,6 +1164,8 @@ event_wait( void )
 	case 0:
 		return;
 	case -1:
+		if (errno == EINTR)
+			return;
 		perror( "select() failed in event loop" );
 		abort();
 	default:
